"""Test pass module."""

import jaclang.compiler.absyntree as ast
from jaclang.compiler.compile import jac_file_to_pass
from jaclang.compiler.passes.main import DeclImplMatchPass
from jaclang.utils.test import TestCase


class DeclImplMatchPassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_parameter_count_mismatch(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            self.fixture_abs_path("defn_decl_mismatch.jac"), DeclImplMatchPass
        )

        expected_stdout_values = (
            "Parameter count mismatch for ability (o)SomeObj.(c)foo.",
            "    8 |",
            "    9 | # Miss match parameter count.",
            "   10 | :obj:SomeObj:can:foo(param1: str) -> str {",
            "      | ^^^^^^^^^^^^^^^^^^^^",
            '   11 |     return "foo";',
            "   12 | }",
            "From the declaration of foo.",
            "    2 |",
            "    3 | obj SomeObj {",
            "    4 |     can foo(param1: str, param2:int) -> str;",
            "      |         ^^^",
            "    5 |     can bar(param1: str, param2:int) -> str;",
            "    6 | }",
        )

        errors_output = ""
        for error in state.errors_had:
            errors_output += error.pretty_print() + "\n"

        for exp in expected_stdout_values:
            self.assertIn(exp, errors_output)

    def test_ability_connected_to_decl(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(self.fixture_abs_path("base.jac"), DeclImplMatchPass)
        self.assertFalse(state.errors_had)
        self.assertIn("(o)Test.(c)say_hi", state.ir.sym_tab.tab)
        self.assertIsNotNone(
            state.ir.sym_tab.tab["(o)Test.(c)say_hi"].decl.name_of.body
        )
        self.assertIn("(o)Test.(c)__init__", state.ir.sym_tab.tab)
        self.assertIsNotNone(
            state.ir.sym_tab.tab["(o)Test.(c)__init__"].decl.name_of.body
        )

    def test_ability_connected_to_decl_post(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(self.fixture_abs_path("base2.jac"), DeclImplMatchPass)
        self.assertFalse(state.errors_had)
        self.assertIn("(o)Test.(c)say_hi", state.ir.sym_tab.tab)
        self.assertIsNotNone(
            state.ir.sym_tab.tab["(o)Test.(c)say_hi"].decl.name_of.body
        )
        self.assertIn("(o)Test.(c)__init__", state.ir.sym_tab.tab)
        self.assertIsNotNone(
            state.ir.sym_tab.tab["(o)Test.(c)__init__"].decl.name_of.body
        )

    def test_arch_ref_has_sym(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            self.fixture_abs_path("defs_and_uses.jac"), DeclImplMatchPass
        )
        for i in state.ir.get_all_sub_nodes(ast.ArchRef):
            self.assertIsNotNone(i.sym)

    def test_obj_hasvar_initialization(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            self.fixture_abs_path("uninitialized_hasvars.jac"), DeclImplMatchPass
        )
        self.assertTrue(state.errors_had)

        expected_stdout_values = (
            "Non default attribute 'var3' follows default attribute",
            "    4 |     has var1: int;",
            "    5 |     has var2: int = 42;",
            "    6 |     has var3: int; # <-- This should be syntax error.",
            "      |         ^^^^",
            "    7 | }",
            'Missing "postinit" method required by un initialized attribute(s).',
            "   11 | obj Test2 {",
            "   12 |     has var1: str;",
            "   13 |     has var2: int by postinit;",
            "      |         ^^^^",
            "   14 | }",
            "Non default attribute 'var4' follows default attribute",
            "   19 |     has var2: int = 42;",
            "   20 |     has var3: int by postinit;  # <-- This is fine.",
            "   21 |     has var4: int;  # <-- This should be syntax error.",
            "      |         ^^^^",
            "   22 |",
            "   23 |     can postinit() {",
        )

        errors_output = ""
        for error in state.errors_had:
            errors_output += error.pretty_print() + "\n"

        for exp in expected_stdout_values:
            self.assertIn(exp, errors_output)



"""Semantic Token Manager module."""

from __future__ import annotations

from typing import List, Optional, Tuple

import jaclang.compiler.absyntree as ast
from jaclang.langserve.utils import (
    find_surrounding_tokens,
    get_line_of_code,
    get_token_start,
)

import lsprotocol.types as lspt


class SemTokManager:
    """Semantic Token Manager class."""

    def __init__(self, ir: ast.Module) -> None:
        """Initialize semantic token manager."""
        self.sem_tokens: List[int] = self.gen_sem_tokens(ir)
        self.static_sem_tokens: List[
            Tuple[lspt.Position, int, int, ast.AstSymbolNode]
        ] = self.gen_sem_tok_node(ir)

    def gen_sem_tokens(self, ir: ast.Module) -> list[int]:
        """Return semantic tokens."""
        tokens = []
        prev_line, prev_col = 0, 0
        for node in ir._in_mod_nodes:
            if isinstance(node, ast.NameAtom) and node.sem_token:
                line, col_start, col_end = (
                    node.loc.first_line - 1,
                    node.loc.col_start - 1,
                    node.loc.col_end - 1,
                )
                length = col_end - col_start
                tokens += [
                    line - prev_line,
                    col_start if line != prev_line else col_start - prev_col,
                    length,
                    *node.sem_token,
                ]
                prev_line, prev_col = line, col_start
        return tokens

    def gen_sem_tok_node(
        self, ir: ast.Module
    ) -> List[Tuple[lspt.Position, int, int, ast.AstSymbolNode]]:
        """Return semantic tokens."""
        tokens: List[Tuple[lspt.Position, int, int, ast.AstSymbolNode]] = []
        for node in ir._in_mod_nodes:
            if isinstance(node, ast.NameAtom) and node.sem_token:
                line, col_start, col_end = (
                    node.loc.first_line - 1,
                    node.loc.col_start - 1,
                    node.loc.col_end - 1,
                )
                length = col_end - col_start
                pos = lspt.Position(line, col_start)
                tokens += [(pos, col_end, length, node)]
        return tokens

    def update_sem_tokens(
        self,
        content_changes: lspt.DidChangeTextDocumentParams,
        sem_tokens: list[int],
        document_lines: List[str],
    ) -> list[int]:
        """Update semantic tokens on change."""
        for change in [
            x
            for x in content_changes.content_changes
            if isinstance(x, lspt.TextDocumentContentChangeEvent_Type1)
        ]:
            change_start_line = change.range.start.line
            change_start_char = change.range.start.character
            change_end_line = change.range.end.line
            change_end_char = change.range.end.character

            is_delete = change.text == ""
            prev_token_index, next_token_index, insert_inside_token = (
                find_surrounding_tokens(
                    change_start_line,
                    change_start_char,
                    change_end_line,
                    change_end_char,
                    sem_tokens,
                )
            )
            prev_tok_pos = get_token_start(prev_token_index, sem_tokens)
            nxt_tok_pos = get_token_start(next_token_index, sem_tokens)
            changing_line_text = get_line_of_code(change_start_line, document_lines)
            if not changing_line_text:
                return sem_tokens
            is_edit_between_tokens = bool(
                (
                    change_start_line > prev_tok_pos[0]
                    or (
                        change_start_line == prev_tok_pos[0]
                        and change_start_char
                        > prev_tok_pos[1] + sem_tokens[prev_token_index + 2]
                        if prev_token_index and prev_token_index + 2 < len(sem_tokens)
                        else 0
                    )
                )
                and (
                    change_end_line < nxt_tok_pos[0]
                    or (
                        change_end_line == nxt_tok_pos[0]
                        and change_end_char < nxt_tok_pos[1]
                    )
                )
            )
            text = r"%s" % change.text
            line_delta = len(text.split("\n")) - 1

            is_multiline_insertion = line_delta > 0
            is_next_token_same_line = change_end_line == nxt_tok_pos[0]

            if is_delete:
                next_token_index = (
                    prev_token_index + 5
                    if insert_inside_token
                    and prev_token_index is not None
                    or (
                        next_token_index
                        and prev_token_index is not None
                        and next_token_index >= 10
                        and next_token_index - prev_token_index == 10
                    )
                    else next_token_index
                )
                if next_token_index is None:
                    return sem_tokens
                nxt_tok_pos = get_token_start(next_token_index, sem_tokens)
                is_single_line_change = change_end_line == change_start_line
                is_next_token_same_line = change_end_line == nxt_tok_pos[0]
                if (
                    is_single_line_change
                    and insert_inside_token
                    and prev_token_index is not None
                ):
                    sem_tokens = SemTokManager.handle_single_line_delete(
                        sem_tokens,
                        next_token_index,
                        prev_token_index,
                        is_next_token_same_line,
                        change,
                    )
                elif is_single_line_change and is_edit_between_tokens:
                    sem_tokens = SemTokManager.handle_single_line_delete_between_tokens(
                        sem_tokens,
                        next_token_index,
                        is_next_token_same_line,
                        change,
                        change_start_line,
                        change_end_line,
                    )
                else:
                    sem_tokens = SemTokManager.handle_multi_line_delete(
                        sem_tokens,
                        next_token_index,
                        nxt_tok_pos,
                        change_start_line,
                        change_end_line,
                        change_start_char,
                        change_end_char,
                        prev_tok_pos,
                        is_next_token_same_line,
                    )
                return sem_tokens

            is_token_boundary_edit = False
            if insert_inside_token and prev_token_index is not None:
                sem_tokens, is_token_boundary_edit, nxt_tok_pos, next_token_index = (
                    SemTokManager.handle_insert_inside_token(
                        change,
                        sem_tokens,
                        prev_token_index,
                        changing_line_text,
                        line_delta,
                        prev_tok_pos,
                        change_start_char,
                        change_end_char,
                        is_token_boundary_edit,
                        nxt_tok_pos,
                    )
                )
            tokens_on_same_line = prev_tok_pos[0] == nxt_tok_pos[0]
            if (
                is_edit_between_tokens
                or is_token_boundary_edit
                or is_multiline_insertion
            ) and next_token_index is not None:
                if is_multiline_insertion:
                    sem_tokens = SemTokManager.handle_multi_line_insertion(
                        sem_tokens,
                        next_token_index,
                        nxt_tok_pos,
                        change_start_line,
                        change_end_line,
                        change_end_char,
                        prev_tok_pos,
                        tokens_on_same_line,
                        changing_line_text,
                        line_delta,
                    )
                else:
                    sem_tokens = SemTokManager.handle_single_line_insertion(
                        sem_tokens,
                        next_token_index,
                        is_next_token_same_line,
                        change,
                        tokens_on_same_line,
                        nxt_tok_pos,
                        change_start_line,
                        line_delta,
                    )
        return sem_tokens

    @staticmethod
    def handle_multi_line_delete(
        sem_tokens: list[int],
        next_token_index: int,
        nxt_tok_pos: tuple[int, int, int],
        change_start_line: int,
        change_end_line: int,
        change_start_char: int,
        change_end_char: int,
        prev_tok_pos: tuple[int, int, int],
        is_next_token_same_line: bool,
    ) -> list[int]:
        """Handle multi line deletion."""
        if is_next_token_same_line:
            char_del = nxt_tok_pos[1] - change_end_char
            total_char_del = change_start_char + char_del
            sem_tokens[next_token_index + 1] = (
                (total_char_del - prev_tok_pos[1])
                if prev_tok_pos[0] == change_start_line
                else total_char_del
            )
        sem_tokens[next_token_index] -= change_end_line - change_start_line
        return sem_tokens

    @staticmethod
    def handle_single_line_delete_between_tokens(
        sem_tokens: list[int],
        next_token_index: int,
        is_next_token_same_line: bool,
        change: lspt.TextDocumentContentChangeEvent_Type1,
        change_start_line: int,
        change_end_line: int,
    ) -> list[int]:
        """Handle single line deletion between tokens."""
        if is_next_token_same_line:
            sem_tokens[next_token_index + 1] -= change.range_length

        else:
            sem_tokens[next_token_index] -= change_end_line - change_start_line
        return sem_tokens

    @staticmethod
    def handle_single_line_delete(
        sem_tokens: list[int],
        next_token_index: int,
        prev_token_index: int,
        is_next_token_same_line: bool,
        change: lspt.TextDocumentContentChangeEvent_Type1,
    ) -> list[int]:
        """Handle single line deletion."""
        sem_tokens[prev_token_index + 2] -= change.range_length
        if is_next_token_same_line:
            sem_tokens[next_token_index + 1] -= change.range_length
        return sem_tokens

    @staticmethod
    def handle_single_line_insertion(
        sem_tokens: list[int],
        next_token_index: int,
        is_next_token_same_line: bool,
        change: lspt.TextDocumentContentChangeEvent_Type1,
        tokens_on_same_line: bool,
        nxt_tok_pos: tuple[int, int, int],
        change_start_line: int,
        line_delta: int,
    ) -> list[int]:
        """Handle single line insertion."""
        if tokens_on_same_line:
            sem_tokens[next_token_index + 1] += len(change.text)
            sem_tokens[next_token_index] += line_delta
        else:
            is_next_token_same_line = change_start_line == nxt_tok_pos[0]
            if is_next_token_same_line:
                sem_tokens[next_token_index] += line_delta
                sem_tokens[next_token_index + 1] += len(change.text)
            else:
                sem_tokens[next_token_index] += line_delta
        return sem_tokens

    @staticmethod
    def handle_multi_line_insertion(
        sem_tokens: list[int],
        next_token_index: int,
        nxt_tok_pos: tuple[int, int, int],
        change_start_line: int,
        change_end_line: int,
        change_end_char: int,
        prev_tok_pos: tuple[int, int, int],
        tokens_on_same_line: bool,
        changing_line_text: tuple[str, int],
        line_delta: int,
    ) -> list[int]:
        """Handle multi line insertion."""
        if tokens_on_same_line:
            char_del = nxt_tok_pos[1] - change_end_char
            total_char_del = changing_line_text[1] + char_del

        else:
            is_prev_token_same_line = change_end_line == prev_tok_pos[0]
            is_next_token_same_line = change_start_line == nxt_tok_pos[0]
            if is_prev_token_same_line:
                total_char_del = nxt_tok_pos[1]
            elif is_next_token_same_line:
                char_del = nxt_tok_pos[1] - change_end_char
                total_char_del = changing_line_text[1] + char_del
            else:
                total_char_del = sem_tokens[next_token_index + 1]
                line_delta -= change_end_line - change_start_line
        sem_tokens[next_token_index + 1] = total_char_del
        sem_tokens[next_token_index] += line_delta
        return sem_tokens

    @staticmethod
    def handle_insert_inside_token(
        change: lspt.TextDocumentContentChangeEvent_Type1,
        sem_tokens: list[int],
        prev_token_index: int,
        changing_line_text: tuple[str, int],
        line_delta: int,
        prev_tok_pos: tuple[int, int, int],
        change_start_char: int,
        change_end_char: int,
        is_token_boundary_edit: bool,
        nxt_tok_pos: tuple[int, int, int],
    ) -> tuple[list[int], bool, tuple[int, int, int], Optional[int]]:
        """Handle insert inside token."""
        next_token_index = None
        for i in ["\n", " ", "\t"]:
            if i in change.text:
                if prev_tok_pos[1] == change_start_char:
                    if i == "\n":
                        sem_tokens[prev_token_index] += line_delta
                        sem_tokens[prev_token_index + 1] = changing_line_text[1]
                    else:
                        sem_tokens[prev_token_index + 1] += len(change.text)
                    return (
                        sem_tokens,
                        is_token_boundary_edit,
                        nxt_tok_pos,
                        next_token_index,
                    )
                else:
                    is_token_boundary_edit = True
                    next_token_index = prev_token_index + 5
                    nxt_tok_pos = get_token_start(next_token_index, sem_tokens)
                    break
        if not is_token_boundary_edit:
            selected_region = change_end_char - change_start_char
            index_offset = 2
            sem_tokens[prev_token_index + index_offset] += (
                len(change.text) - selected_region
            )
            if prev_tok_pos[0] == get_token_start(prev_token_index + 5, sem_tokens)[0]:
                sem_tokens[prev_token_index + index_offset + 4] += (
                    len(change.text) - selected_region
                )
        return sem_tokens, is_token_boundary_edit, nxt_tok_pos, next_token_index



"""Subnode Table building pass.

This pass builds a table of subnodes for each node in the AST. This is used
for fast lookup of nodes of a certain type in the AST. This is just a utility
pass and is not required for any other pass to work.
"""

from copy import copy

import jaclang.compiler.absyntree as ast
from jaclang.compiler.passes import Pass


class SubNodeTabPass(Pass):
    """AST Enrichment Pass for basic high level semantics."""

    def enter_node(self, node: ast.AstNode) -> None:
        """Table builder."""
        super().enter_node(node)
        node._sub_node_tab = {}  # clears on entry

    def exit_node(self, node: ast.AstNode) -> None:
        """Table builder."""
        super().exit_node(node)
        for i in node.kid:
            if not i:
                continue
            for k, v in i._sub_node_tab.items():
                if k in node._sub_node_tab:
                    node._sub_node_tab[k].extend(v)
                else:
                    node._sub_node_tab[k] = copy(v)
            if type(i) in node._sub_node_tab:
                node._sub_node_tab[type(i)].append(i)
            else:
                node._sub_node_tab[type(i)] = [i]



"""Jac Symbol Table."""

from __future__ import annotations

import ast as ast3
from typing import Optional, Sequence

import jaclang.compiler.absyntree as ast
from jaclang.compiler.constant import SymbolAccess, SymbolType
from jaclang.utils.treeprinter import dotgen_symtab_tree, print_symtab_tree


# Symbols can have mulitple definitions but resolves decl to be the
# first such definition in a given scope.
class Symbol:
    """Symbol."""

    def __init__(
        self,
        defn: ast.NameAtom,
        access: SymbolAccess,
        parent_tab: SymbolTable,
    ) -> None:
        """Initialize."""
        self.defn: list[ast.NameAtom] = [defn]
        self.uses: list[ast.NameAtom] = []
        defn.sym = self
        self.access: SymbolAccess = access
        self.parent_tab = parent_tab

    @property
    def decl(self) -> ast.NameAtom:
        """Get decl."""
        return self.defn[0]

    @property
    def sym_name(self) -> str:
        """Get name."""
        return self.decl.sym_name

    @property
    def sym_type(self) -> SymbolType:
        """Get sym_type."""
        return self.decl.sym_category

    @property
    def sym_dotted_name(self) -> str:
        """Return a full path of the symbol."""
        out = [self.defn[0].sym_name]
        current_tab: SymbolTable | None = self.parent_tab
        while current_tab is not None:
            out.append(current_tab.name)
            current_tab = current_tab.parent
        out.reverse()
        return ".".join(out)

    @property
    def fetch_sym_tab(self) -> Optional[SymbolTable]:
        """Get symbol table."""
        return self.parent_tab.find_scope(self.sym_name)

    def add_defn(self, node: ast.NameAtom) -> None:
        """Add defn."""
        self.defn.append(node)
        node.sym = self

    def add_use(self, node: ast.NameAtom) -> None:
        """Add use."""
        self.uses.append(node)
        node.sym = self

    def __repr__(self) -> str:
        """Repr."""
        return f"Symbol({self.sym_name}, {self.sym_type}, {self.access}, {self.defn})"


class SymbolTable:
    """Symbol Table."""

    def __init__(
        self, name: str, owner: ast.AstNode, parent: Optional[SymbolTable] = None
    ) -> None:
        """Initialize."""
        self.name = name
        self.owner = owner
        self.parent = parent
        self.kid: list[SymbolTable] = []
        self.tab: dict[str, Symbol] = {}
        self.inherit: list[SymbolTable] = []

    def get_type(self) -> SymbolType:
        """Get type."""
        if isinstance(self.owner, ast.AstSymbolNode):
            return self.owner.sym_category
        return SymbolType.VAR

    def get_parent(self) -> Optional[SymbolTable]:
        """Get parent."""
        return self.parent

    def lookup(self, name: str, deep: bool = True) -> Optional[Symbol]:
        """Lookup a variable in the symbol table."""
        if name in self.tab:
            return self.tab[name]
        for i in self.inherit:
            found = i.lookup(name, deep=False)
            if found:
                return found
        if deep and self.parent:
            return self.parent.lookup(name, deep)
        return None

    def insert(
        self,
        node: ast.AstSymbolNode,
        access_spec: Optional[ast.AstAccessNode] | SymbolAccess = None,
        single: bool = False,
        force_overwrite: bool = False,
    ) -> Optional[ast.AstNode]:
        """Set a variable in the symbol table.

        Returns original symbol as collision if single check fails, none otherwise.
        Also updates node.sym to create pointer to symbol.
        """
        collision = (
            self.tab[node.sym_name].defn[-1]
            if single and node.sym_name in self.tab
            else None
        )
        if force_overwrite or node.sym_name not in self.tab:
            self.tab[node.sym_name] = Symbol(
                defn=node.name_spec,
                access=(
                    access_spec
                    if isinstance(access_spec, SymbolAccess)
                    else access_spec.access_type if access_spec else SymbolAccess.PUBLIC
                ),
                parent_tab=self,
            )
        else:
            self.tab[node.sym_name].add_defn(node.name_spec)
        node.name_spec.sym = self.tab[node.sym_name]
        return collision

    def find_scope(self, name: str) -> Optional[SymbolTable]:
        """Find a scope in the symbol table."""
        for k in self.kid:
            if k.name == name:
                return k
        return None

    def push_kid_scope(self, name: str, key_node: ast.AstNode) -> SymbolTable:
        """Push a new scope onto the symbol table."""
        self.kid.append(SymbolTable(name, key_node, self))
        return self.kid[-1]

    def inherit_sym_tab(self, target_sym_tab: SymbolTable) -> None:
        """Inherit symbol table."""
        for i in target_sym_tab.tab.values():
            self.def_insert(i.decl, access_spec=i.access)

    def def_insert(
        self,
        node: ast.AstSymbolNode,
        access_spec: Optional[ast.AstAccessNode] | SymbolAccess = None,
        single_decl: Optional[str] = None,
        force_overwrite: bool = False,
    ) -> Optional[Symbol]:
        """Insert into symbol table."""
        if node.sym and self == node.sym.parent_tab:
            return node.sym
        self.insert(
            node=node,
            single=single_decl is not None,
            access_spec=access_spec,
            force_overwrite=force_overwrite,
        )
        self.update_py_ctx_for_def(node)
        return node.sym

    def chain_def_insert(self, node_list: list[ast.AstSymbolNode]) -> None:
        """Link chain of containing names to symbol."""
        if not node_list:
            return
        cur_sym_tab: SymbolTable | None = node_list[0].sym_tab
        node_list[-1].name_spec.py_ctx_func = ast3.Store
        if isinstance(node_list[-1].name_spec, ast.AstSymbolNode):
            node_list[-1].name_spec.py_ctx_func = ast3.Store

        node_list = node_list[:-1]  # Just performs lookup mappings of pre assign chain
        for i in node_list:
            cur_sym_tab = (
                lookup.decl.sym_tab
                if (
                    lookup := self.use_lookup(
                        i,
                        sym_table=cur_sym_tab,
                    )
                )
                else None
            )

    def use_lookup(
        self,
        node: ast.AstSymbolNode,
        sym_table: Optional[SymbolTable] = None,
    ) -> Optional[Symbol]:
        """Link to symbol."""
        if node.sym:
            return node.sym
        if not sym_table:
            sym_table = node.sym_tab
        if sym_table:
            lookup = sym_table.lookup(name=node.sym_name, deep=True)
            lookup.add_use(node.name_spec) if lookup else None
        return node.sym

    def chain_use_lookup(self, node_list: Sequence[ast.AstSymbolNode]) -> None:
        """Link chain of containing names to symbol."""
        if not node_list:
            return
        cur_sym_tab: SymbolTable | None = node_list[0].sym_tab
        for i in node_list:
            if cur_sym_tab is None:
                break
            lookup = self.use_lookup(i, sym_table=cur_sym_tab)
            if lookup:
                cur_sym_tab = lookup.decl.sym_tab

                # check if the symbol table name is not the same as symbol name
                # then try to find a child scope with the same name
                # This is used to get the scope in case of
                #      import:py math;
                #      b = math.floor(1.7);
                if cur_sym_tab.name != i.sym_name:
                    t = cur_sym_tab.find_scope(i.sym_name)
                    if t:
                        cur_sym_tab = t
            else:
                cur_sym_tab = None

    def update_py_ctx_for_def(self, node: ast.AstSymbolNode) -> None:
        """Update python context for definition."""
        node.name_spec.py_ctx_func = ast3.Store
        if isinstance(node, (ast.TupleVal, ast.ListVal)) and node.values:
            # Handling of UnaryExpr case for item is only necessary for
            # the generation of Starred nodes in the AST for examples
            # like `(a, *b) = (1, 2, 3, 4)`.
            def fix(item: ast.TupleVal | ast.ListVal | ast.UnaryExpr) -> None:
                if isinstance(item, ast.UnaryExpr):
                    if isinstance(item.operand, ast.AstSymbolNode):
                        item.operand.name_spec.py_ctx_func = ast3.Store
                elif isinstance(item, (ast.TupleVal, ast.ListVal)):
                    for i in item.values.items if item.values else []:
                        if isinstance(i, ast.AstSymbolNode):
                            i.name_spec.py_ctx_func = ast3.Store
                        elif isinstance(i, ast.AtomTrailer):
                            self.chain_def_insert(i.as_attr_list)
                        if isinstance(i, (ast.TupleVal, ast.ListVal, ast.UnaryExpr)):
                            fix(i)

            fix(node)

    def inherit_baseclasses_sym(self, node: ast.Architype | ast.Enum) -> None:
        """Inherit base classes symbol tables."""
        if node.base_classes:
            for base_cls in node.base_classes.items:
                if (
                    isinstance(base_cls, ast.AstSymbolNode)
                    and (found := self.use_lookup(base_cls))
                    and found
                ):
                    self.inherit.append(found.decl.sym_tab)
                    base_cls.name_spec.name_of = found.decl.name_of

    def pp(self, depth: Optional[int] = None) -> str:
        """Pretty print."""
        return print_symtab_tree(root=self, depth=depth)

    def dotgen(self) -> str:
        """Generate dot graph for sym table."""
        return dotgen_symtab_tree(self)

    def __repr__(self) -> str:
        """Repr."""
        out = f"{self.name} {super().__repr__()}:\n"
        for k, v in self.tab.items():
            out += f"    {k}: {v}\n"
        return out


__all__ = [
    "Symbol",
    "SymbolTable",
    "SymbolType",
    "SymbolAccess",
]



"""Connect Decls and Defs in AST.

This pass creates and manages compilation of Python code from the AST. This pass
also creates bytecode files from the Python code, and manages the caching of
relevant files.
"""

import os


import jaclang.compiler.absyntree as ast
from jaclang.compiler.constant import Constants as Con
from jaclang.compiler.passes import Pass


class PyOutPass(Pass):
    """Python and bytecode file printing pass."""

    def before_pass(self) -> None:
        """Before pass."""
        return super().before_pass()

    def enter_module(self, node: ast.Module) -> None:
        """Sub objects.

        name: str,
        doc: Token,
        body: Optional['Elements'],
        mod_path: str,
        is_imported: bool,
        sym_tab: Optional[SymbolTable],
        """
        if not (os.path.exists(node.loc.mod_path) and node.gen.py_ast):
            self.error(
                f"Unable to find module {node.loc.mod_path} or no code present.", node
            )
            return
        mods = [node] + [
            i for i in self.get_all_sub_nodes(node, ast.Module) if not i.stub_only
        ]
        for mod in mods:
            mod_path, out_path_py, out_path_pyc = self.get_output_targets(mod)
            if os.path.exists(out_path_pyc) and os.path.getmtime(
                out_path_pyc
            ) > os.path.getmtime(mod_path):
                continue
            try:
                self.gen_python(mod, out_path=out_path_py)
                self.dump_bytecode(mod, mod_path=mod_path, out_path=out_path_pyc)
            except Exception as e:
                self.warning(f"Error in generating Python code: {e}", node)
        self.terminate()

    def gen_python(self, node: ast.Module, out_path: str) -> None:
        """Generate Python."""
        with open(out_path, "w") as f:
            f.write(node.gen.py)

    def dump_bytecode(self, node: ast.Module, mod_path: str, out_path: str) -> None:
        """Generate Python."""
        if node.gen.py_bytecode:
            with open(out_path, "wb") as f:
                f.write(node.gen.py_bytecode)
        else:
            self.error(
                f"Soemthing went wrong with {node.loc.mod_path} compilation.", node
            )

    def get_output_targets(self, node: ast.Module) -> tuple[str, str, str]:
        """Get output targets."""
        base_path, file_name = os.path.split(node.loc.mod_path)
        gen_path = os.path.join(base_path, Con.JAC_GEN_DIR)
        mod_dir, file_name = os.path.split(node.loc.mod_path)
        mod_dir = mod_dir.replace(base_path, "").lstrip(os.sep)
        base_name, _ = os.path.splitext(file_name)
        out_dir = os.path.join(gen_path, mod_dir)
        try:
            os.makedirs(gen_path, exist_ok=True)
            with open(os.path.join(gen_path, "__init__.py"), "w"):
                pass
            os.makedirs(out_dir, exist_ok=True)
        except Exception as e:
            self.warning(f"Can't create directory {out_dir}: {e}", node)
        out_path_py = os.path.join(out_dir, f"{base_name}.py")
        out_path_pyc = os.path.join(out_dir, f"{base_name}.jbc")
        return node.loc.mod_path, out_path_py, out_path_pyc



"""Access check pass for the Jac compiler.

This pass checks for access to attributes in the Jac language.
"""

from typing import Optional

import jaclang.compiler.absyntree as ast
from jaclang.compiler.constant import SymbolAccess
from jaclang.compiler.passes import Pass
from jaclang.compiler.symtable import Symbol
from jaclang.settings import settings


class AccessCheckPass(Pass):
    """Jac Ast Access Check pass."""

    # NOTE: This method is a hacky way to detect if the drivied class is inherit from base class, it
    # doesn't work if the base class was provided as an expression (ex. obj Dri :module.Base: {...}).
    def is_class_inherited_from(
        self, dri_class: ast.Architype, base_class: ast.Architype
    ) -> bool:
        """Return true if the dri_class inherited from base_class."""
        if dri_class.base_classes is None:
            return False
        for expr in dri_class.base_classes.items:
            if not isinstance(expr, ast.Name):
                continue
            if not isinstance(expr.name_of, ast.Architype):
                continue  # Unlikely.
            if expr.name_of == base_class:
                return True
            if self.is_class_inherited_from(expr.name_of, base_class):
                return True
        return False

    def report_error(self, message: str, node: Optional[ast.AstNode] = None) -> None:
        """Report error message related to illegal access of attributes and objects."""
        self.error(message, node)

    def exit_node(self, node: ast.AstNode) -> None:  # TODO: Move to debug pass
        """Exit node."""
        super().exit_node(node)
        if (
            settings.lsp_debug
            and isinstance(node, ast.NameAtom)
            and not node.sym
            and not node.parent_of_type(ast.Module).py_info.is_raised_from_py
            and not (
                node.sym_name == "py"
                and node.parent
                and isinstance(node.parent.parent, ast.Import)
            )
        ):
            self.warning(f"Name {node.sym_name} not present in symbol table")

    def enter_name(self, node: ast.Name) -> None:
        """Sub objects.

        name: str,
        value: str,
        col_start: int,
        col_end: int,
        pos_start: int,
        pos_end: int,
        """
        # TODO: Enums are not considered at the moment, I'll need to test and add them bellow.

        # If the current node is a global variable's name there is no access, it's just the declaration.
        if Pass.find_parent_of_type(node, ast.GlobalVars) is not None:
            return

        # Class name, and ability name are declarations and there is no access here as well.
        if isinstance(node.name_of, (ast.Ability, ast.Architype, ast.Enum)):
            return

        # Get the context to check the access.
        curr_class: Optional[ast.Architype] = Pass.find_parent_of_type(
            node, ast.Architype
        )
        curr_module: Optional[ast.Module] = Pass.find_parent_of_type(node, ast.Module)
        if curr_module is None:
            return

        # Note that currently we can only check for name + symbols, because expressions are not associated with the
        # typeinfo thus they don't have a symbol. In the future the name nodes will become expression nodes.
        if not isinstance(node.sym, Symbol):
            return

        # Public symbols are fine.
        if node.sym.access == SymbolAccess.PUBLIC:
            return

        # Note that from bellow the access is either private or protected.
        is_portect = node.sym.access == SymbolAccess.PROTECTED
        access_type = "protected" if is_portect else "private"

        # The class we're currently in (None if we're not inside any).
        sym_owner: ast.AstNode = node.sym.parent_tab.owner

        # If the symbol belongs to a class, we need to check if the access used properly
        # within the class and in it's inherited classes.
        if isinstance(sym_owner, ast.Architype):

            # Accessing a private/protected member within the top level scope illegal.
            if curr_class is None:
                return self.report_error(
                    f'Error: Invalid access of {access_type} member "{node.sym_name}".',
                    node,
                )

            if curr_class != node.sym.parent_tab.owner:
                if not is_portect:  # private member accessed in a different class.
                    return self.report_error(
                        f'Error: Invalid access of {access_type} member "{node.sym_name}".',
                        node,
                    )
                else:  # Accessing a protected member, check we're in an inherited class.
                    if not self.is_class_inherited_from(curr_class, sym_owner):
                        return self.report_error(
                            f'Error: Invalid access of {access_type} member "{node.sym_name}".',
                            node,
                        )

        elif isinstance(sym_owner, ast.Module) and sym_owner != curr_module:
            # Accessing a private/public member in a different module.
            return self.report_error(
                f'Error: Invalid access of {access_type} member "{node.sym_name}".',
                node,
            )



"""Utility functions for the language server."""

import asyncio
import builtins
import re
from functools import wraps
from typing import Any, Awaitable, Callable, Coroutine, Optional, ParamSpec, TypeVar

import jaclang.compiler.absyntree as ast
from jaclang.compiler.codeloc import CodeLocInfo
from jaclang.compiler.constant import SymbolType
from jaclang.compiler.passes.transform import Alert
from jaclang.compiler.symtable import Symbol, SymbolTable
from jaclang.vendor.pygls import uris

import lsprotocol.types as lspt

T = TypeVar("T", bound=Callable[..., Coroutine[Any, Any, Any]])
P = ParamSpec("P")


def gen_diagnostics(
    from_path: str, errors: list[Alert], warnings: list[Alert]
) -> list[lspt.Diagnostic]:
    """Return diagnostics."""
    return [
        lspt.Diagnostic(
            range=create_range(error.loc),
            message=error.msg,
            severity=lspt.DiagnosticSeverity.Error,
        )
        for error in errors
        if error.loc.mod_path == uris.to_fs_path(from_path)
    ] + [
        lspt.Diagnostic(
            range=create_range(warning.loc),
            message=warning.msg,
            severity=lspt.DiagnosticSeverity.Warning,
        )
        for warning in warnings
        if warning.loc.mod_path == uris.to_fs_path(from_path)
    ]


def debounce(wait: float) -> Callable[[T], Callable[..., Awaitable[None]]]:
    """Debounce decorator for async functions."""

    def decorator(fn: T) -> Callable[..., Awaitable[None]]:
        @wraps(fn)
        async def debounced(*args: P.args, **kwargs: P.kwargs) -> None:
            async def call_it() -> None:
                await fn(*args, **kwargs)

            if hasattr(debounced, "_task"):
                debounced._task.cancel()

            async def debounced_coro() -> None:
                try:
                    await asyncio.sleep(wait)
                    await call_it()
                except asyncio.CancelledError:
                    pass

            setattr(  # noqa: B010
                debounced, "_task", asyncio.create_task(debounced_coro())
            )

        return debounced

    return decorator


def sym_tab_list(sym_tab: SymbolTable, file_path: str) -> list[SymbolTable]:
    """Iterate through symbol table."""
    sym_tabs = (
        [sym_tab]
        if not (
            isinstance(sym_tab.owner, ast.Module)
            and sym_tab.owner.loc.mod_path != file_path
        )
        else []
    )
    for i in sym_tab.kid:
        sym_tabs += sym_tab_list(i, file_path=file_path)
    return sym_tabs


def find_deepest_symbol_node_at_pos(
    node: ast.AstNode, line: int, character: int
) -> Optional[ast.AstSymbolNode]:
    """Return the deepest symbol node that contains the given position."""
    last_symbol_node = None

    if position_within_node(node, line, character):
        if isinstance(node, ast.AstSymbolNode):
            last_symbol_node = node

        for child in [i for i in node.kid if i.loc.mod_path == node.loc.mod_path]:
            if position_within_node(child, line, character):
                deeper_node = find_deepest_symbol_node_at_pos(child, line, character)
                if deeper_node is not None:
                    last_symbol_node = deeper_node

    return last_symbol_node


def position_within_node(node: ast.AstNode, line: int, character: int) -> bool:
    """Check if the position falls within the node's location."""
    if node.loc.first_line < line + 1 < node.loc.last_line:
        return True
    if (
        node.loc.first_line == line + 1
        and node.loc.col_start <= character + 1
        and (
            node.loc.last_line == line + 1
            and node.loc.col_end >= character + 1
            or node.loc.last_line > line + 1
        )
    ):
        return True
    if (
        node.loc.last_line == line + 1
        and node.loc.col_start <= character + 1 <= node.loc.col_end
    ):
        return True
    return False


def find_index(
    sem_tokens: list[int],
    line: int,
    char: int,
) -> Optional[int]:
    """Find index."""
    index = None

    # A list contains all the token start positions.
    token_start_list = [
        get_token_start(i, sem_tokens) for i in range(0, len(sem_tokens), 5)
    ]
    for i, j in enumerate(token_start_list):
        if j[0] == line and j[1] <= char <= j[2]:
            return i

    return index


def get_symbols_for_outline(node: SymbolTable) -> list[lspt.DocumentSymbol]:
    """Recursively collect symbols from the AST."""
    symbols = []
    for key, item in node.tab.items():
        if (
            key in dir(builtins)
            or item in [owner_sym(tab) for tab in node.kid]
            or item.decl.loc.mod_path != node.owner.loc.mod_path
        ):
            continue
        pos = create_range(item.decl.loc)
        symbol = lspt.DocumentSymbol(
            name=key,
            kind=kind_map(item.decl),
            range=pos,
            selection_range=pos,
            children=[],
        )
        symbols.append(symbol)

    for sub_tab in [
        i for i in node.kid if i.owner.loc.mod_path == node.owner.loc.mod_path
    ]:
        sub_symbols = get_symbols_for_outline(sub_tab)
        if isinstance(
            sub_tab.owner,
            (ast.IfStmt, ast.ElseStmt, ast.WhileStmt, ast.IterForStmt, ast.InForStmt),
        ):
            symbols.extend(sub_symbols)
        else:
            sub_pos = create_range(sub_tab.owner.loc)
            symbol = lspt.DocumentSymbol(
                name=sub_tab.name,
                kind=kind_map(sub_tab.owner),
                range=sub_pos,
                selection_range=sub_pos,
                children=sub_symbols,
            )
            symbols.append(symbol)

    return symbols


def owner_sym(table: SymbolTable) -> Optional[Symbol]:
    """Get owner sym."""
    if table.parent and isinstance(table.owner, ast.AstSymbolNode):
        return table.parent.lookup(table.owner.sym_name)
    return None


def create_range(loc: CodeLocInfo) -> lspt.Range:
    """Create an lspt.Range from a location object."""
    return lspt.Range(
        start=lspt.Position(
            line=loc.first_line - 1 if loc.first_line > 0 else 0,
            character=loc.col_start - 1 if loc.col_start > 0 else 0,
        ),
        end=lspt.Position(
            line=loc.last_line - 1 if loc.last_line > 0 else 0,
            character=loc.col_end - 1 if loc.col_end > 0 else 0,
        ),
    )


def get_location_range(mod_item: ast.ModuleItem) -> tuple[int, int, int, int]:
    """Get location range."""
    if not mod_item.from_mod_path.sub_module:
        raise ValueError("Module items should have module path. Not Possible.")
    lookup = mod_item.from_mod_path.sub_module.sym_tab.lookup(mod_item.name.value)
    if not lookup:
        raise ValueError("Module items should have a symbol table entry. Not Possible.")
    loc = lookup.decl.loc
    return loc.first_line, loc.col_start, loc.last_line, loc.col_end


def kind_map(sub_tab: ast.AstNode) -> lspt.SymbolKind:
    """Map the symbol node to an lspt.SymbolKind."""
    return (
        lspt.SymbolKind.Function
        if isinstance(sub_tab, (ast.Ability, ast.AbilityDef))
        else (
            lspt.SymbolKind.Class
            if isinstance(sub_tab, (ast.Architype, ast.ArchDef))
            else (
                lspt.SymbolKind.Module
                if isinstance(sub_tab, ast.Module)
                else (
                    lspt.SymbolKind.Enum
                    if isinstance(sub_tab, (ast.Enum, ast.EnumDef))
                    else lspt.SymbolKind.Variable
                )
            )
        )
    )


def label_map(sub_tab: SymbolType) -> lspt.CompletionItemKind:
    """Map the symbol node to an lspt.CompletionItemKind."""
    return (
        lspt.CompletionItemKind.Function
        if sub_tab in [SymbolType.ABILITY, SymbolType.TEST]
        else (
            lspt.CompletionItemKind.Class
            if sub_tab
            in [
                SymbolType.OBJECT_ARCH,
                SymbolType.NODE_ARCH,
                SymbolType.EDGE_ARCH,
                SymbolType.WALKER_ARCH,
            ]
            else (
                lspt.CompletionItemKind.Module
                if sub_tab == SymbolType.MODULE
                else (
                    lspt.CompletionItemKind.Enum
                    if sub_tab == SymbolType.ENUM_ARCH
                    else (
                        lspt.CompletionItemKind.Field
                        if sub_tab == SymbolType.HAS_VAR
                        else (
                            lspt.CompletionItemKind.Method
                            if sub_tab == SymbolType.METHOD
                            else (
                                lspt.CompletionItemKind.EnumMember
                                if sub_tab == SymbolType.ENUM_MEMBER
                                else (
                                    lspt.CompletionItemKind.Interface
                                    if sub_tab == SymbolType.IMPL
                                    else lspt.CompletionItemKind.Variable
                                )
                            )
                        )
                    )
                )
            )
        )
    )


def collect_all_symbols_in_scope(
    sym_tab: SymbolTable, up_tree: bool = True
) -> list[lspt.CompletionItem]:
    """Return all symbols in scope."""
    symbols = []
    visited = set()
    current_tab: Optional[SymbolTable] = sym_tab

    while current_tab is not None and current_tab not in visited:
        visited.add(current_tab)
        for name, symbol in current_tab.tab.items():
            if name not in dir(builtins) and symbol.sym_type != SymbolType.IMPL:
                symbols.append(
                    lspt.CompletionItem(label=name, kind=label_map(symbol.sym_type))
                )
        if not up_tree:
            return symbols
        current_tab = current_tab.parent if current_tab.parent != current_tab else None
    return symbols


def collect_child_tabs(sym_tab: SymbolTable) -> list[lspt.CompletionItem]:
    """Return all child tab's as completion items."""
    symbols: list[lspt.CompletionItem] = []
    for tab in sym_tab.kid:
        if tab.name not in [i.label for i in symbols]:
            symbols.append(
                lspt.CompletionItem(label=tab.name, kind=label_map(tab.get_type()))
            )
    return symbols


def parse_symbol_path(text: str, dot_position: int) -> list[str]:
    """Parse text and return a list of symbols."""
    text = text[:dot_position][:-1].strip()
    valid_character_pattern = re.compile(r"[a-zA-Z0-9_]")

    reversed_text = text[::-1]
    all_words = []
    current_word = []
    for char in reversed_text:
        if valid_character_pattern.fullmatch(char):
            current_word.append(char)
        elif char == ".":
            if current_word:
                all_words.append("".join(current_word[::-1]))
                current_word = []
        else:
            if current_word:
                all_words.append("".join(current_word[::-1]))
                current_word = []
            break

    all_words = (
        all_words[::-1]
        if not current_word
        else ["".join(current_word[::-1])] + all_words[::-1]
    )

    return all_words


def get_token_start(
    token_index: int | None, sem_tokens: list[int]
) -> tuple[int, int, int]:
    """Return the starting position of a token."""
    if token_index is None or token_index >= len(sem_tokens):
        return 0, 0, 0

    current_line = 0
    current_char = 0
    current_tok_index = 0

    while current_tok_index < len(sem_tokens):
        token_line_delta = sem_tokens[current_tok_index]
        token_start_char = sem_tokens[current_tok_index + 1]

        if token_line_delta > 0:
            current_line += token_line_delta
            current_char = 0
        if current_tok_index == token_index:
            if token_line_delta > 0:
                return (
                    current_line,
                    token_start_char,
                    token_start_char + sem_tokens[current_tok_index + 2],
                )
            return (
                current_line,
                current_char + token_start_char,
                current_char + token_start_char + sem_tokens[current_tok_index + 2],
            )

        current_char += token_start_char
        current_tok_index += 5

    return (
        current_line,
        current_char,
        current_char + sem_tokens[current_tok_index + 2],
    )


def find_surrounding_tokens(
    change_start_line: int,
    change_start_char: int,
    change_end_line: int,
    change_end_char: int,
    sem_tokens: list[int],
) -> tuple[int | None, int | None, bool]:
    """Find the indices of the previous and next tokens surrounding the change."""
    prev_token_index = None
    next_token_index = None
    inside_tok = False
    for i, tok in enumerate(
        [get_token_start(i, sem_tokens) for i in range(0, len(sem_tokens), 5)][0:]
    ):
        if (not (prev_token_index is None or next_token_index is None)) and (
            tok[0] > change_end_line
            or (tok[0] == change_end_line and tok[1] > change_end_char)
        ):
            prev_token_index = i * 5
            break
        elif (
            change_start_line == tok[0] == change_end_line
            and tok[1] <= change_start_char
            and tok[2] >= change_end_char
        ):
            prev_token_index = i * 5
            inside_tok = True
            break
        elif (tok[0] < change_start_line) or (
            tok[0] == change_start_line and tok[1] < change_start_char
        ):
            prev_token_index = i * 5
        elif (tok[0] > change_end_line) or (
            tok[0] == change_end_line and tok[1] > change_end_char
        ):
            next_token_index = i * 5
            break

    return prev_token_index, next_token_index, inside_tok


def get_line_of_code(line_number: int, lines: list[str]) -> Optional[tuple[str, int]]:
    """Get the line of code, and the first non-space character index."""
    if 0 <= line_number < len(lines):
        line = lines[line_number].rstrip("\n")
        first_non_space = len(line) - len(line.lstrip())
        return line, (
            first_non_space + 4
            if line.strip().endswith(("(", "{", "["))
            else first_non_space
        )
    return None


def add_unique_text_edit(
    changes: dict[str, list[lspt.TextEdit]], key: str, new_edit: lspt.TextEdit
) -> None:
    """Add a new text edit to the changes dictionary if it is unique."""
    if key not in changes:
        changes[key] = [new_edit]
    else:
        for existing_edit in changes[key]:
            if (
                existing_edit.range.start == new_edit.range.start
                and existing_edit.range.end == new_edit.range.end
                and existing_edit.new_text == new_edit.new_text
            ):
                return
        changes[key].append(new_edit)



"""This is a helper file."""


def add1():
    return 1 + 1


def sub1():
    return 1 - 1


# classes
class Orange1:
    def __init__(self):
        self.orange = 1

    def get_orange(self):
        return self.orange

    def set_orange(self, orange):
        self.orange = orange


# variables
orange2 = 1
apple = 1



"""Collect Python dependencies based on reference sourced from Jac.

This pass will use mypy to calculate the python package/module dependencies
that are only relevant to actual references source from Jac code.
"""

from __future__ import annotations


import os

import jaclang.compiler.absyntree as ast
from jaclang.compiler.passes import Pass
from jaclang.settings import settings

import mypy.nodes as MypyNodes  # noqa N812


class PyCollectDepsPass(Pass):
    """Python and bytecode file self.__debug_printing pass."""

    def __debug_print(self, msg: str) -> None:
        if settings.collect_py_dep_debug:
            self.log_info("CollectPythonDependencies::" + msg)

    def enter_node(self, node: ast.AstNode) -> None:
        """Collect python dependencies from all Jac Nodes."""
        assert isinstance(self.ir, ast.Module)

        if not isinstance(node, ast.AstSymbolNode):
            return

        # Adding the path of the file related to the py import
        path: str = ""
        if isinstance(node, ast.ModulePath):
            if node.path:
                path = ".".join([i.value for i in node.path])
            node.abs_path = self.ir.py_info.py_mod_dep_map.get(path)
            if node.abs_path and os.path.isfile(node.abs_path.replace(".pyi", ".py")):
                node.abs_path = node.abs_path.replace(".pyi", ".py")

        elif isinstance(node, ast.ModuleItem):
            imp = node.parent_of_type(ast.Import)
            mod_path_node = imp.get_all_sub_nodes(ast.ModulePath)[0]
            if mod_path_node.path:
                path = ".".join([i.value for i in mod_path_node.path])
            path += f".{node.name.value}"
            node.abs_path = self.ir.py_info.py_mod_dep_map.get(path)
            if node.abs_path and os.path.isfile(node.abs_path.replace(".pyi", ".py")):
                node.abs_path = node.abs_path.replace(".pyi", ".py")

        if len(node.gen.mypy_ast) == 0:
            return

        mypy_node = node.gen.mypy_ast[0]

        if isinstance(mypy_node, MypyNodes.RefExpr) and mypy_node.node:
            node_full_name = mypy_node.node.fullname
            if "." in node_full_name:
                mod_name = node_full_name[: node_full_name.rindex(".")]
            else:
                mod_name = node_full_name

            if mod_name not in self.ir.py_info.py_mod_dep_map:
                self.__debug_print(
                    f"Can't find a python file associated with {type(node)}::{node.loc}"
                )
                return

            mode_path = self.ir.py_info.py_mod_dep_map[mod_name]
            if mode_path.endswith(".jac"):
                return

            self.ir.py_info.py_raise_map[mod_name] = mode_path
        else:
            self.__debug_print(
                f"Collect python dependencies is not supported in {type(node)}::{node.loc}"
            )



"""Test pass module."""

from typing import List

from jaclang.compiler.compile import jac_file_to_pass
from jaclang.compiler.passes.main.schedules import py_code_gen_typed
from jaclang.utils.lang_tools import AstTool
from jaclang.utils.test import TestCase


class MypyTypeCheckPassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        self.__messages: List[str] = []
        return super().setUp()

    def test_type_errors(self) -> None:
        """Basic test for pass."""
        type_checked = jac_file_to_pass(
            file_path=self.fixture_abs_path("func.jac"),
            schedule=py_code_gen_typed,
        )

        errs = "\n".join([i.msg for i in type_checked.warnings_had])
        files = "\n".join([i.loc.mod_path for i in type_checked.warnings_had])

        for i in [
            "func2.jac",
            "func.jac",
            '(got "int", expected "str")',
            '(got "str", expected "int")',
        ]:
            self.assertIn(i, errs + files)

    def test_imported_module_typecheck(self) -> None:
        """Basic test for pass."""
        type_checked = jac_file_to_pass(
            file_path=self.fixture_abs_path("game1.jac"),
            schedule=py_code_gen_typed,
        )

        errs = "\n".join([i.msg for i in type_checked.warnings_had])
        files = "\n".join([i.loc.mod_path for i in type_checked.warnings_had])

        for i in [
            'Argument 2 to "is_pressed" of "Button" has incompatible type "int"; expected "str"',
        ]:
            self.assertIn(i, errs + files)

    def test_type_coverage(self) -> None:
        """Testing for type info coverage in sym_tab via ast."""
        out = AstTool().ir(["ast", f"{self.fixture_abs_path('type_info.jac')}"])
        lis = [
            "introduce - Type: None",
            "25:30 - 25:37\t        |   |   |           +-- Name - species - Type: None",  # AtomTrailer
        ]
        self.assertIn("HasVar - species - Type: builtins.str", out)
        self.assertIn("myDog - Type: type_info.Dog", out)
        self.assertIn("Body - Type: type_info.Dog.Body", out)
        self.assertEqual(out.count("Type: builtins.str"), 35)
        for i in lis:
            self.assertNotIn(i, out)

    def test_data_spatial_type_info(self) -> None:
        """Testing for type info for dataspatial constructs."""
        out = AstTool().ir(
            ["ast", f"{self.fixture_abs_path('data_spatial_types.jac')}"]
        )
        self.assertRegex(
            out,
            r"129:24 - 129:28.*SpecialVarRef - _Jac.get_root\(\) \- Type\: jaclang.runtimelib.architype.Root",
        )
        self.assertRegex(out, r"129:11 - 129:29.*FuncCall \- Type\: builtins\.str")
        self.assertRegex(
            out,
            r"129:15 - 129:23.*Name \- node_dot \- Type\: builtins.str,  SymbolTable\: str",
        )
        self.assertRegex(
            out,
            r"128:5 - 128:25.*BinaryExpr \- Type\: jaclang.runtimelib.architype.WalkerArchitype",
        )
        self.assertRegex(
            out,
            r"48:11 - 48:28.*EdgeRefTrailer \- Type\: builtins.list\[data_spatial_types.A\]",
        )
        self.assertRegex(out, r"24:5 - 24:25.*BinaryExpr \- Type\: builtins.bool", out)



"""The Jac Programming Language."""

from jaclang.plugin.default import JacFeatureImpl
from jaclang.plugin.feature import JacFeature, plugin_manager

jac_import = JacFeature.jac_import


plugin_manager.register(JacFeatureImpl)
plugin_manager.load_setuptools_entrypoints("jac")

__all__ = ["jac_import"]



"""Jaclang Language Server."""

from __future__ import annotations

from typing import Optional

from jaclang.compiler.constant import (
    JacSemTokenModifier as SemTokMod,
    JacSemTokenType as SemTokType,
)
from jaclang.langserve.engine import JacLangServer
from jaclang.settings import settings

import lsprotocol.types as lspt

server = JacLangServer()


@server.feature(lspt.TEXT_DOCUMENT_DID_OPEN)
async def did_open(ls: JacLangServer, params: lspt.DidOpenTextDocumentParams) -> None:
    """Check syntax on change."""
    ls.deep_check(params.text_document.uri)
    ls.lsp.send_request(lspt.WORKSPACE_SEMANTIC_TOKENS_REFRESH)


@server.feature(lspt.TEXT_DOCUMENT_DID_SAVE)
async def did_save(ls: JacLangServer, params: lspt.DidOpenTextDocumentParams) -> None:
    """Check syntax on change."""
    file_path = params.text_document.uri
    if ls.modules[file_path].is_modified:
        await ls.launch_deep_check(file_path)
        ls.lsp.send_request(lspt.WORKSPACE_SEMANTIC_TOKENS_REFRESH)
        ls.modules[file_path].is_modified = False


@server.feature(lspt.TEXT_DOCUMENT_DID_CHANGE)
async def did_change(
    ls: JacLangServer, params: lspt.DidChangeTextDocumentParams
) -> None:
    """Check syntax on change."""
    module = ls.modules[params.text_document.uri]
    module.is_modified = True
    await ls.launch_quick_check(file_path := params.text_document.uri)
    if file_path in ls.modules:
        document = ls.workspace.get_text_document(file_path)
        lines = document.source.splitlines()
        ls.modules[file_path].sem_manager.update_sem_tokens(
            params, ls.modules[file_path].sem_manager.sem_tokens, lines
        )
        ls.lsp.send_request(lspt.WORKSPACE_SEMANTIC_TOKENS_REFRESH)


@server.feature(lspt.TEXT_DOCUMENT_FORMATTING)
def formatting(
    ls: JacLangServer, params: lspt.DocumentFormattingParams
) -> list[lspt.TextEdit]:
    """Format the given document."""
    return ls.formatted_jac(params.text_document.uri)


@server.feature(
    lspt.WORKSPACE_DID_CREATE_FILES,
    lspt.FileOperationRegistrationOptions(
        filters=[
            lspt.FileOperationFilter(pattern=lspt.FileOperationPattern("**/*.jac"))
        ]
    ),
)
def did_create_files(ls: JacLangServer, params: lspt.CreateFilesParams) -> None:
    """Check syntax on file creation."""


@server.feature(
    lspt.WORKSPACE_DID_RENAME_FILES,
    lspt.FileOperationRegistrationOptions(
        filters=[
            lspt.FileOperationFilter(pattern=lspt.FileOperationPattern("**/*.jac"))
        ]
    ),
)
def did_rename_files(ls: JacLangServer, params: lspt.RenameFilesParams) -> None:
    """Check syntax on file rename."""
    new_uris = [file.new_uri for file in params.files]
    old_uris = [file.old_uri for file in params.files]
    for i in range(len(new_uris)):
        ls.rename_module(old_uris[i], new_uris[i])


@server.feature(
    lspt.WORKSPACE_DID_DELETE_FILES,
    lspt.FileOperationRegistrationOptions(
        filters=[
            lspt.FileOperationFilter(pattern=lspt.FileOperationPattern("**/*.jac"))
        ]
    ),
)
def did_delete_files(ls: JacLangServer, params: lspt.DeleteFilesParams) -> None:
    """Check syntax on file delete."""
    for file in params.files:
        ls.delete_module(file.uri)


@server.feature(
    lspt.TEXT_DOCUMENT_COMPLETION,
    lspt.CompletionOptions(trigger_characters=[".", ":", "a-zA-Z0-9"]),
)
def completion(ls: JacLangServer, params: lspt.CompletionParams) -> lspt.CompletionList:
    """Provide completion."""
    return ls.get_completion(
        params.text_document.uri,
        params.position,
        params.context.trigger_character if params.context else None,
    )


@server.feature(lspt.TEXT_DOCUMENT_HOVER, lspt.HoverOptions(work_done_progress=True))
def hover(
    ls: JacLangServer, params: lspt.TextDocumentPositionParams
) -> Optional[lspt.Hover]:
    """Provide hover information for the given hover request."""
    return ls.get_hover_info(params.text_document.uri, params.position)


@server.feature(lspt.TEXT_DOCUMENT_DOCUMENT_SYMBOL)
def document_symbol(
    ls: JacLangServer, params: lspt.DocumentSymbolParams
) -> list[lspt.DocumentSymbol]:
    """Provide document symbols."""
    return ls.get_outline(params.text_document.uri)


@server.feature(lspt.TEXT_DOCUMENT_DEFINITION)
def definition(
    ls: JacLangServer, params: lspt.TextDocumentPositionParams
) -> Optional[lspt.Location]:
    """Provide definition."""
    return ls.get_definition(params.text_document.uri, params.position)


@server.feature(lspt.TEXT_DOCUMENT_REFERENCES)
def references(ls: JacLangServer, params: lspt.ReferenceParams) -> list[lspt.Location]:
    """Provide references."""
    return ls.get_references(params.text_document.uri, params.position)


@server.feature(lspt.TEXT_DOCUMENT_RENAME)
def rename(
    ls: JacLangServer, params: lspt.RenameParams
) -> Optional[lspt.WorkspaceEdit]:
    """Rename symbol."""
    ls.log_warning("Auto Rename is Experimental, Please use with caution.")
    return ls.rename_symbol(params.text_document.uri, params.position, params.new_name)


@server.feature(
    lspt.TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
    lspt.SemanticTokensLegend(
        token_types=SemTokType.as_str_list(),
        token_modifiers=SemTokMod.as_str_list(),
    ),
)
def semantic_tokens_full(
    ls: JacLangServer, params: lspt.SemanticTokensParams
) -> lspt.SemanticTokens:
    """Provide semantic tokens."""
    return ls.get_semantic_tokens(params.text_document.uri)


def run_lang_server() -> None:
    """Run the language server."""
    settings.pass_timer = True
    server.start_io()


if __name__ == "__main__":
    run_lang_server()



"""Integrate mypy infrastructure into Jac.

This is used to call mypy type checking into Jac files by integrating
mypy apis into Jac and use jac py ast in it.
"""

from __future__ import annotations

import re
from typing import Callable, Optional, TypeVar

import jaclang.compiler.absyntree as ast
from jaclang.compiler.constant import Constants, Tokens
from jaclang.compiler.passes import Pass
from jaclang.compiler.symtable import SymbolTable
from jaclang.settings import settings
from jaclang.utils.helpers import pascal_to_snake
from jaclang.vendor.mypy.nodes import Node as VNode  # bit of a hack


import mypy.nodes as MypyNodes  # noqa N812
import mypy.types as MypyTypes  # noqa N812
from mypy.checkexpr import Type as MyType


T = TypeVar("T", bound=ast.AstSymbolNode)


class FuseTypeInfoPass(Pass):
    """Python and bytecode file self.__debug_printing pass."""

    node_type_hash: dict[MypyNodes.Node | VNode, MyType] = {}

    # Override this to support enter expression.
    def enter_node(self, node: ast.AstNode) -> None:
        """Run on entering node."""
        super().enter_node(node)

        if isinstance(node, ast.Expr):
            self.enter_expr(node)

    def __debug_print(self, msg: str) -> None:
        if settings.fuse_type_info_debug:
            self.log_info("FuseTypeInfo::" + msg)

    def __call_type_handler(self, mypy_type: MypyTypes.Type) -> Optional[str]:
        mypy_type_name = pascal_to_snake(mypy_type.__class__.__name__)
        type_handler_name = f"get_type_from_{mypy_type_name}"
        if hasattr(self, type_handler_name):
            return getattr(self, type_handler_name)(mypy_type)
        self.__debug_print(
            f'"MypyTypes::{mypy_type.__class__.__name__}" isn\'t supported yet'
        )
        return None

    def __set_type_sym_table_link(self, node: ast.AstSymbolNode) -> None:
        typ = node.expr_type.split(".")
        typ_sym_table = self.ir.sym_tab
        assert isinstance(self.ir, ast.Module)

        sym_type = node.expr_type
        if re.match(r"builtins.(list|dict|tuple)", sym_type):
            sym_type = re.sub(r"\[.*\]", "", sym_type)

        typ = sym_type.split(".")

        if node.expr_type == "types.ModuleType" and node.sym:
            node.name_spec.type_sym_tab = node.sym.parent_tab.find_scope(node.sym_name)

        for i in typ:
            if i == self.ir.name:
                continue
            f = typ_sym_table.find_scope(i)
            if f:
                typ_sym_table = f

        if typ_sym_table != self.ir.sym_tab:
            node.name_spec.type_sym_tab = typ_sym_table

    @staticmethod
    def __handle_node(
        func: Callable[[FuseTypeInfoPass, T], None]
    ) -> Callable[[FuseTypeInfoPass, T], None]:
        def node_handler(self: FuseTypeInfoPass, node: T) -> None:
            if not isinstance(node, ast.AstSymbolNode):
                self.__debug_print(
                    f"Warning {node.__class__.__name__} is not an AstSymbolNode"
                )

            try:
                jac_node_str = f'jac node "{node.loc}::{node.__class__.__name__}'
                if hasattr(node, "value"):
                    jac_node_str += f'::{node.value}"'
                else:
                    jac_node_str += '"'

                # Jac node has only one mypy node linked to it
                if len(node.gen.mypy_ast) == 1:
                    func(self, node)
                    self.__set_type_sym_table_link(node)

                # Jac node has multiple mypy nodes linked to it
                elif len(node.gen.mypy_ast) > 1:
                    # Checking that these nodes are duplicate or not
                    temp = []
                    for n in node.gen.mypy_ast:
                        n_id = f"{type(node)}"
                        n_id += f"::{n.line}::{n.column}"
                        n_id += f" - {n.end_line}::{n.end_column}"
                        if n_id not in temp:
                            temp.append(n_id)

                    # Check the number of unique mypy nodes linked
                    if len(temp) > 1:
                        self.__debug_print(
                            f"{jac_node_str} has multiple mypy nodes associated to it"
                        )
                    else:
                        self.__debug_print(
                            f"{jac_node_str} has duplicate mypy nodes associated to it"
                        )
                        func(self, node)
                        self.__set_type_sym_table_link(node)

                # Special handing for BuiltinType
                elif isinstance(node, ast.BuiltinType):
                    func(self, node)  # type: ignore
                    self.__set_type_sym_table_link(node)

                # Jac node doesn't have mypy nodes linked to it
                else:
                    self.__debug_print(
                        f"{jac_node_str} doesn't have mypy node associated to it"
                    )

            except AttributeError as e:
                if settings.fuse_type_info_debug:
                    raise e
                self.__debug_print(
                    f'{node.loc}: Internal error happened while parsing "{e.obj.__class__.__name__}"'
                )

        return node_handler

    def __collect_type_from_symbol(self, node: ast.AstSymbolNode) -> None:
        mypy_node = node.gen.mypy_ast[0]

        if isinstance(mypy_node, MypyNodes.MemberExpr):
            if mypy_node in self.node_type_hash:
                node.name_spec.expr_type = (
                    self.__call_type_handler(self.node_type_hash[mypy_node])
                    or node.name_spec.expr_type
                )
            else:
                self.__debug_print(f"{node.loc} MemberExpr type is not found")

        elif hasattr(mypy_node, "node"):
            # orig_node = mypy_node
            mypy_node = mypy_node.node

            if isinstance(mypy_node, (MypyNodes.Var, MypyNodes.FuncDef)):
                node.name_spec.expr_type = (
                    self.__call_type_handler(mypy_node.type) or node.name_spec.expr_type
                )

            elif isinstance(mypy_node, MypyNodes.MypyFile):
                node.name_spec.expr_type = "types.ModuleType"

            elif isinstance(mypy_node, MypyNodes.TypeInfo):
                node.name_spec.expr_type = mypy_node.fullname

            elif isinstance(mypy_node, MypyNodes.OverloadedFuncDef):
                node.name_spec.expr_type = (
                    self.__call_type_handler(mypy_node.items[0].func.type)
                    or node.name_spec.expr_type
                )

            elif mypy_node is None:
                node.name_spec.expr_type = "None"

            else:
                self.__debug_print(
                    f'"{node.loc}::{node.__class__.__name__}" mypy (with node attr) node isn\'t supported'
                    f"{type(mypy_node)}"
                )

        else:
            if isinstance(mypy_node, MypyNodes.ClassDef):
                node.name_spec.expr_type = mypy_node.fullname
                self.__set_type_sym_table_link(node)
            elif isinstance(mypy_node, MypyNodes.FuncDef):
                node.name_spec.expr_type = (
                    self.__call_type_handler(mypy_node.type) or node.name_spec.expr_type
                )
            elif isinstance(mypy_node, MypyNodes.Argument):
                node.name_spec.expr_type = (
                    self.__call_type_handler(mypy_node.variable.type)
                    or node.name_spec.expr_type
                )
            elif isinstance(mypy_node, MypyNodes.Decorator):
                node.name_spec.expr_type = (
                    self.__call_type_handler(mypy_node.func.type.ret_type)
                    or node.name_spec.expr_type
                )
            else:
                self.__debug_print(
                    f'"{node.loc}::{node.__class__.__name__}" mypy node isn\'t supported'
                    f"{type(mypy_node)}"
                )

    def __check_builltin_symbol(self, node: ast.NameAtom) -> None:
        if isinstance(node.parent, ast.AtomTrailer) and node is node.parent.right:
            return
        builtins_sym_tab = self.ir.sym_tab.find_scope("builtins")
        assert builtins_sym_tab is not None
        builtins_sym = builtins_sym_tab.lookup(node.sym_name)
        if builtins_sym:
            node.name_spec._sym = builtins_sym

    collection_types_map = {
        ast.ListVal: "builtins.list",
        ast.SetVal: "builtins.set",
        ast.TupleVal: "builtins.tuple",
        ast.DictVal: "builtins.dict",
        ast.ListCompr: None,
        ast.DictCompr: None,
    }

    # NOTE (Thakee): Since expression nodes are not AstSymbolNodes, I'm not decorating this with __handle_node
    # and IMO instead of checking if it's a symbol node or an expression, we somehow mark expressions as
    # valid nodes that can have symbols. At this point I'm leaving this like this and lemme know
    # otherwise.
    # NOTE (GAMAL): This will be fixed through the AstTypedNode
    def enter_expr(self: FuseTypeInfoPass, node: ast.Expr) -> None:
        """Enter an expression node."""
        if len(node.gen.mypy_ast) == 0:
            return

        # No need to run this handling in case of a previous type
        # was set during the pass, if not then we hope that the
        # mypy node has a type associated to it
        if node.expr_type != "NoType":
            return

        # Check if the expression is a data spatial expression
        # Support disconnectOp
        if isinstance(node, ast.BinaryExpr):
            if isinstance(node.op, ast.DisconnectOp):
                node.expr_type = "builtins.bool"
                return

            # Support spwan and connectOp
            elif (
                isinstance(node.op, ast.ConnectOp)
                or node.op.name == Tokens.KW_SPAWN.value
            ):
                if node.gen.mypy_ast[-1] in self.node_type_hash:
                    node.expr_type = (
                        self.__call_type_handler(
                            self.node_type_hash[node.gen.mypy_ast[-1]]
                        )
                        or node.expr_type
                    )
                return

        if isinstance(node, ast.EdgeRefTrailer) and any(
            isinstance(k, ast.FilterCompr) for k in node.kid
        ):
            if node.gen.mypy_ast[-1] in self.node_type_hash:
                node.expr_type = (
                    self.__call_type_handler(self.node_type_hash[node.gen.mypy_ast[-1]])
                    or node.expr_type
                )
            return

        # If the corrosponding mypy ast node type has stored here, get the values.
        mypy_node = node.gen.mypy_ast[0]
        if mypy_node in self.node_type_hash:
            mytype: MyType = self.node_type_hash[mypy_node]
            node.expr_type = self.__call_type_handler(mytype) or node.expr_type

        # Set they symbol type for collection expression.
        #
        # GenCompr is an instance of ListCompr but we don't handle it here.
        # so the isinstace (node, <classes>) doesn't work, I'm going with type(...) == ...
        if type(node) in self.collection_types_map:
            assert isinstance(node, ast.AtomExpr)  # To make mypy happy.
            collection_type = self.collection_types_map[type(node)]
            if collection_type is not None:
                node.name_spec.expr_type = collection_type
            if mypy_node in self.node_type_hash:
                node.name_spec.expr_type = (
                    self.__call_type_handler(mytype) or node.name_spec.expr_type
                )

    @__handle_node
    def enter_name(self, node: ast.NameAtom) -> None:
        """Pass handler for name nodes."""
        self.__collect_type_from_symbol(node)
        if node.sym is None:
            self.__check_builltin_symbol(node)

    @__handle_node
    def enter_module_path(self, node: ast.ModulePath) -> None:
        """Pass handler for ModulePath nodes."""
        self.__debug_print(f"Getting type not supported in {type(node)}")

    @__handle_node
    def enter_module_item(self, node: ast.ModuleItem) -> None:
        """Pass handler for ModuleItem nodes."""
        self.__debug_print(f"Getting type not supported in {type(node)}")

    @__handle_node
    def enter_architype(self, node: ast.Architype) -> None:
        """Pass handler for Architype nodes."""
        self.__collect_type_from_symbol(node)

    @__handle_node
    def enter_arch_def(self, node: ast.ArchDef) -> None:
        """Pass handler for ArchDef nodes."""
        self.__debug_print(f"Getting type not supported in {type(node)}")

    @__handle_node
    def enter_enum(self, node: ast.Enum) -> None:
        """Pass handler for Enum nodes."""
        self.__collect_type_from_symbol(node)

    @__handle_node
    def enter_enum_def(self, node: ast.EnumDef) -> None:
        """Pass handler for EnumDef nodes."""
        self.__debug_print(f"Getting type not supported in {type(node)}")

    @__handle_node
    def enter_ability(self, node: ast.Ability) -> None:
        """Pass handler for Ability nodes."""
        if isinstance(node.gen.mypy_ast[0], MypyNodes.FuncDef):
            node.name_spec.expr_type = (
                self.__call_type_handler(node.gen.mypy_ast[0].type.ret_type)
                or node.name_spec.expr_type
            )
        else:
            self.__debug_print(
                f"{node.loc}: Can't get type of an ability from mypy node other than Ability. "
                f"{type(node.gen.mypy_ast[0])}"
            )

    @__handle_node
    def enter_ability_def(self, node: ast.AbilityDef) -> None:
        """Pass handler for AbilityDef nodes."""
        if isinstance(node.gen.mypy_ast[0], MypyNodes.FuncDef):
            node.name_spec.expr_type = (
                self.__call_type_handler(node.gen.mypy_ast[0].type.ret_type)
                or node.name_spec.expr_type
            )
        else:
            self.__debug_print(
                f"{node.loc}: Can't get type of an AbilityDef from mypy node other than FuncDef. "
                f"{type(node.gen.mypy_ast[0])}"
            )

    @__handle_node
    def enter_param_var(self, node: ast.ParamVar) -> None:
        """Pass handler for ParamVar nodes."""
        if isinstance(node.gen.mypy_ast[0], MypyNodes.Argument):
            mypy_node: MypyNodes.Argument = node.gen.mypy_ast[0]
            if mypy_node.variable.type:
                node.name_spec.expr_type = (
                    self.__call_type_handler(mypy_node.variable.type)
                    or node.name_spec.expr_type
                )
        else:
            self.__debug_print(
                f"{node.loc}: Can't get parameter value from mypyNode other than Argument"
            )

    # TODO: support all lhs if needed
    @__handle_node
    def enter_has_var(self, node: ast.HasVar) -> None:
        """Pass handler for HasVar nodes."""
        mypy_node = node.gen.mypy_ast[0]
        if isinstance(mypy_node, MypyNodes.AssignmentStmt):
            n = mypy_node.lvalues[0].node
            if isinstance(n, (MypyNodes.Var, MypyNodes.FuncDef)):
                node.name_spec.expr_type = (
                    self.__call_type_handler(n.type) or node.name_spec.expr_type
                )
            else:
                self.__debug_print(
                    "Getting type of 'AssignmentStmt' is only supported with Var and FuncDef"
                )
        else:
            self.__debug_print(
                "Getting type of 'HasVar' is only supported with AssignmentStmt"
            )

    def exit_has_var(self, node: ast.HasVar) -> None:
        """Pass handler for HasVar nodes."""
        node.name_spec.expr_type = node.name.expr_type
        node.name_spec.type_sym_tab = node.name.type_sym_tab

    @__handle_node
    def enter_multi_string(self, node: ast.MultiString) -> None:
        """Pass handler for MultiString nodes."""
        node.name_spec.expr_type = "builtins.str"

    @__handle_node
    def enter_f_string(self, node: ast.FString) -> None:
        """Pass handler for FString nodes."""
        self.__debug_print(f"Getting type not supported in {type(node)}")

    @__handle_node
    def enter_index_slice(self, node: ast.IndexSlice) -> None:
        """Pass handler for IndexSlice nodes."""
        self.__debug_print(f"Getting type not supported in {type(node)}")

    @__handle_node
    def enter_arch_ref(self, node: ast.ArchRef) -> None:
        """Pass handler for ArchRef nodes."""
        if isinstance(node.gen.mypy_ast[0], MypyNodes.ClassDef):
            mypy_node: MypyNodes.ClassDef = node.gen.mypy_ast[0]
            node.name_spec.expr_type = mypy_node.fullname
            self.__set_type_sym_table_link(node)
        elif isinstance(node.gen.mypy_ast[0], MypyNodes.FuncDef):
            mypy_node2: MypyNodes.FuncDef = node.gen.mypy_ast[0]
            node.name_spec.expr_type = (
                self.__call_type_handler(mypy_node2.type) or node.name_spec.expr_type
            )
        else:
            self.__debug_print(
                f"{node.loc}: Can't get ArchRef value from mypyNode other than ClassDef "
                f"type(node.gen.mypy_ast[0])"
            )

    @__handle_node
    def enter_special_var_ref(self, node: ast.SpecialVarRef) -> None:
        """Pass handler for SpecialVarRef nodes."""
        if node.py_resolve_name() == Constants.ROOT:
            if node.gen.mypy_ast[-1] in self.node_type_hash:
                node.name_spec.expr_type = (
                    self.__call_type_handler(self.node_type_hash[node.gen.mypy_ast[-1]])
                    or node.name_spec.expr_type
                )
        else:
            self.enter_name(node)

    @__handle_node
    def enter_edge_op_ref(self, node: ast.EdgeOpRef) -> None:
        """Pass handler for EdgeOpRef nodes."""
        self.__collect_type_from_symbol(node)

    @__handle_node
    def enter_filter_compr(self, node: ast.FilterCompr) -> None:
        """Pass handler for FilterCompr nodes."""
        self.__debug_print(f"Getting type not supported in {type(node)}")

    @__handle_node
    def enter_assign_compr(self, node: ast.AssignCompr) -> None:
        """Pass handler for AssignCompr nodes."""
        self.__debug_print(f"Getting type not supported in {type(node)}")

    @__handle_node
    def enter_int(self, node: ast.Int) -> None:
        """Pass handler for Int nodes."""
        node.name_spec.expr_type = "builtins.int"

    @__handle_node
    def enter_float(self, node: ast.Float) -> None:
        """Pass handler for Float nodes."""
        node.name_spec.expr_type = "builtins.float"

    @__handle_node
    def enter_string(self, node: ast.String) -> None:
        """Pass handler for String nodes."""
        node.name_spec.expr_type = "builtins.str"

    @__handle_node
    def enter_bool(self, node: ast.Bool) -> None:
        """Pass handler for Bool nodes."""
        node.name_spec.expr_type = "builtins.bool"

    @__handle_node
    def enter_builtin_type(self, node: ast.BuiltinType) -> None:
        """Pass handler for BuiltinType nodes."""
        self.__check_builltin_symbol(node)
        node.name_spec.expr_type = f"builtins.{node.sym_name}"

    def get_type_from_instance(self, mypy_type: MypyTypes.Instance) -> Optional[str]:
        """Get type info from mypy type Instance."""
        #  FIXME: Returning str(mypy_type) won't work for literal values since it would be
        # like Literal['foo'] instead of builtins.str, so we need to get the type fullname.
        # Not sure if this is the best way to do it.
        ret = str(mypy_type)
        if ret.startswith("Literal[") and ret.endswith("]"):
            return mypy_type.type.fullname
        return ret

    def get_type_from_callable_type(
        self, mypy_type: MypyTypes.CallableType
    ) -> Optional[str]:
        """Get type info from mypy type CallableType."""
        return self.__call_type_handler(mypy_type.ret_type)

    # TODO: Which overloaded function to get the return value from?
    def get_type_from_overloaded(
        self, mypy_type: MypyTypes.Overloaded
    ) -> Optional[str]:
        """Get type info from mypy type Overloaded."""
        return self.__call_type_handler(mypy_type.items[-1])

    def get_type_from_none_type(self, mypy_type: MypyTypes.NoneType) -> Optional[str]:
        """Get type info from mypy type NoneType."""
        return "None"

    def get_type_from_any_type(self, mypy_type: MypyTypes.AnyType) -> Optional[str]:
        """Get type info from mypy type NoneType."""
        return "Any"

    def get_type_from_tuple_type(self, mypy_type: MypyTypes.TupleType) -> Optional[str]:
        """Get type info from mypy type TupleType."""
        return "builtins.tuple"

    def get_type_from_type_type(self, mypy_type: MypyTypes.TypeType) -> Optional[str]:
        """Get type info from mypy type TypeType."""
        return str(mypy_type.item)

    def get_type_from_type_var_type(
        self, mypy_type: MypyTypes.TypeVarType
    ) -> Optional[str]:
        """Get type info from mypy type TypeType."""
        return str(mypy_type.name)

    def exit_assignment(self, node: ast.Assignment) -> None:
        """Add new symbols in the symbol table in case of self."""
        # This will fix adding new items to the class through self
        # self.x = 5  # will add x to self datatype symbol table
        for target in node.target.items:
            if (
                isinstance(target, ast.AtomTrailer)
                and isinstance(target.target, ast.SpecialVarRef)
                and target.target.sym_name == "self"
            ):
                self_obj = target.target
                right_obj = target.right
                if self_obj.type_sym_tab and isinstance(right_obj, ast.AstSymbolNode):
                    self_obj.type_sym_tab.def_insert(right_obj)

        # Support adding the correct type symbol table in case of ModuleType
        # This will only work for target=Val & target1=target2=val and
        #   target is a moduleType
        # it won't work in case of
        #   - target1, target2 = Val1, Val2  <-- tuples need more attention
        #   - target = a.b.c                 <-- will be fixed after thakee's PR
        #   - a.b.c = val                    <-- will be fixed after thakee's PR
        for target in node.target.items:
            if (
                isinstance(target, ast.Name)
                and target.expr_type == "types.ModuleType"
                and isinstance(node.value, ast.Name)
            ):
                target.type_sym_tab = node.value.type_sym_tab

    def expand_atom_trailer(self, node_list: list[ast.AstNode]) -> list[ast.AstNode]:
        """Expand the atom trailer object in a list of AstNode."""
        out: list[ast.AstNode] = []
        for i in node_list:
            if isinstance(i, ast.AtomTrailer):
                out.append(i.target)
                out.append(i.right)
            elif isinstance(i, ast.FuncCall):
                out.append(i.target)
            else:
                out.append(i)
        return out

    def exit_atom_trailer(self, node: ast.AtomTrailer) -> None:
        """Adding symbol links to AtomTrailer right nodes."""
        # This will fix adding the symbol links to nodes in atom trailer
        # self.x.z = 5  # will add symbol links to both x and z

        # This function originally used `as_attr_list` in AtomTrailer
        # but an issue happened when doing stuff like fool_me().CONST_VALUE2
        # The issue was due to the way `as_attr_list` implemented so the fix
        # was to implement it again to get all items in AtomTrailer even if
        # their type is not an AstSymbolNode
        atom_trailer_unwind = self.expand_atom_trailer([node])
        iteration_count = 0
        while any(
            isinstance(i, (ast.AtomTrailer, ast.FuncCall)) for i in atom_trailer_unwind
        ):
            atom_trailer_unwind = self.expand_atom_trailer(atom_trailer_unwind)
            iteration_count += 1
            if iteration_count > 50:
                break

        for i in range(1, len(atom_trailer_unwind)):
            left = atom_trailer_unwind[i - 1]
            right = atom_trailer_unwind[i]

            assert isinstance(left, (ast.Expr, ast.AstSymbolNode))
            assert isinstance(right, ast.AstSymbolNode)

            if isinstance(right, ast.IndexSlice):
                # In case of index slice, left won't have a symbol table as it's a list/dict/set
                node_type: str = ""

                # left type is a list
                if left.expr_type.startswith("builtins.list["):
                    node_type = left.expr_type[len("builtins.list[") : -1]

                    # right index slice is a range then it's type is the same as left
                    if right.is_range:
                        right.expr_type = left.expr_type
                        right.parent_of_type(ast.AtomTrailer).expr_type = node_type
                        continue

                # left type is a dictionary
                elif left.expr_type.startswith("builtins.dict["):
                    node_type = (
                        left.expr_type[len("builtins.dict[") : -1].split(",")[1].strip()
                    )

                # unsupported type
                else:
                    continue

                right.expr_type = node_type
                right.parent_of_type(ast.AtomTrailer).expr_type = node_type

                # Getting the correct symbol table and link it
                type_symtab: Optional[SymbolTable] = self.ir.sym_tab
                assert isinstance(self.ir, ast.Module)
                assert isinstance(type_symtab, SymbolTable)

                if re.match(r"builtins.(list|dict|tuple)", node_type):
                    node_type = re.sub(r"\[.*\]", "", node_type)

                for j in node_type.split("."):
                    if j == self.ir.name:
                        continue
                    type_symtab = type_symtab.find_scope(j)
                    if type_symtab is None:
                        break
                right.type_sym_tab = type_symtab

            else:
                # Fix the symbolTable linking in case of type annotations
                if left.type_sym_tab is None and isinstance(node.parent, ast.SubTag):
                    assert isinstance(left, ast.AstSymbolNode)
                    left.name_spec.type_sym_tab = self.ir.sym_tab.find_scope(
                        left.sym_name
                    )

                if left.type_sym_tab:
                    right.name_spec.sym = left.type_sym_tab.lookup(right.sym_name)
                    if right.name_spec.sym:
                        right.name_spec.sym.add_use(right.name_spec)



"""Helper for construct."""

from __future__ import annotations

import ast as ast3
import sys
from contextlib import contextmanager
from types import UnionType
from typing import Callable, Iterator, TYPE_CHECKING

import jaclang.compiler.absyntree as ast
from jaclang.compiler.semtable import SemScope

if TYPE_CHECKING:
    from jaclang.runtimelib.constructs import NodeAnchor, NodeArchitype


@contextmanager
def sys_path_context(path: str) -> Iterator[None]:
    """Add a path to sys.path temporarily."""
    novel_path = path not in sys.path
    try:
        if novel_path:
            sys.path.append(path)
        yield
    finally:
        if novel_path:
            sys.path.remove(path)


def collect_node_connections(
    current_node: NodeAnchor,
    visited_nodes: set,
    connections: set,
) -> None:
    """Nodes and edges representing the graph are collected in visited_nodes and connections."""
    if current_node not in visited_nodes:
        visited_nodes.add(current_node)
        edges = current_node.edges
        for edge_ in edges:
            target = edge_.target
            if target:
                connections.add(
                    (
                        current_node.architype,
                        target.architype,
                        edge_.__class__.__name__,
                    )
                )
                collect_node_connections(target, visited_nodes, connections)


def traverse_graph(
    node: NodeArchitype,
    cur_depth: int,
    depth: int,
    edge_type: list[str],
    traverse: bool,
    connections: list,
    node_depths: dict[NodeArchitype, int],
    visited_nodes: list,
    queue: list,
    bfs: bool,
    dfs: Callable,
    node_limit: int,
    edge_limit: int,
) -> None:
    """Traverse the graph using Breadth-First Search (BFS) or Depth-First Search (DFS)."""
    for edge in node.__jac__.edges:
        is_self_loop = id(edge.source) == id(edge.target)
        is_in_edge = edge.target == node.__jac__
        if (traverse and is_in_edge) or edge.architype.__class__.__name__ in edge_type:
            continue
        if is_self_loop:
            continue  # lets skip self loop for a while, need to handle it later
        elif (other_nda := edge.target if not is_in_edge else edge.source) and (
            other_nd := other_nda.architype
        ):
            new_con = (
                (node, other_nd, edge.architype)
                if not is_in_edge
                else (other_nd, node, edge.architype)
            )
            if node in node_depths and node_depths[node] is not None:
                if other_nd in node_depths:
                    node_depths[node] = min(
                        cur_depth, node_depths[node], node_depths[other_nd] + 1
                    )
                    node_depths[other_nd] = min(
                        cur_depth + 1, node_depths[node] + 1, node_depths[other_nd]
                    )
                else:
                    if other_nd:
                        node_depths[other_nd] = min(
                            cur_depth + 1, node_depths[node] + 1
                        )
                    else:
                        raise ValueError("Edge is detached from node in graph")
            if (
                other_nd
                and new_con not in connections
                and (
                    (
                        depth < 0
                        or min(node_depths[node], node_depths[other_nd]) + 1 <= depth
                    )
                    and node_limit > len(visited_nodes)
                    and edge_limit > len(connections)
                )
            ):
                connections.append(new_con)
                if bfs:
                    queue.append([other_nd, cur_depth + 1])
                else:

                    dfs(other_nd, cur_depth + 1)


def get_sem_scope(node: ast.AstNode) -> SemScope:
    """Get scope of the node."""
    a = (
        node.name
        if isinstance(node, ast.Module)
        else (
            node.name.value
            if isinstance(node, (ast.Enum, ast.Architype))
            else node.name_ref.sym_name if isinstance(node, ast.Ability) else ""
        )
    )
    if isinstance(node, ast.Module):
        return SemScope(a, "Module", None)
    elif isinstance(node, (ast.Enum, ast.Architype, ast.Ability)):
        node_type = (
            node.__class__.__name__
            if isinstance(node, ast.Enum)
            else ("Ability" if isinstance(node, ast.Ability) else node.arch_type.value)
        )
        if node.parent:
            return SemScope(
                a,
                node_type,
                get_sem_scope(node.parent),
            )
    else:
        if node.parent:
            return get_sem_scope(node.parent)
    return SemScope("", "", None)


def extract_type(node: ast.AstNode) -> list[str]:
    """Collect type information in assignment using bfs."""
    extracted_type = []
    if isinstance(node, (ast.BuiltinType, ast.Token)):
        extracted_type.append(node.value)
    for child in node.kid:
        extracted_type.extend(extract_type(child))
    return extracted_type


def extract_params(
    body: ast.FuncCall,
) -> tuple[dict[str, ast.Expr], list[tuple[str, ast3.AST]], list[tuple[str, ast3.AST]]]:
    """Extract model parameters, include and exclude information."""
    model_params = {}
    include_info = []
    exclude_info = []
    if body.params:
        for param in body.params.items:
            if isinstance(param, ast.KWPair) and isinstance(param.key, ast.Name):
                key = param.key.value
                value = param.value
                if key not in ["incl_info", "excl_info"]:
                    model_params[key] = value
                elif key == "incl_info":
                    if isinstance(value, ast.AtomUnit):
                        var_name = (
                            value.value.right.value
                            if isinstance(value.value, ast.AtomTrailer)
                            and isinstance(value.value.right, ast.Name)
                            else (
                                value.value.value
                                if isinstance(value.value, ast.Name)
                                else ""
                            )
                        )
                        include_info.append((var_name, value.gen.py_ast[0]))
                    elif isinstance(value, ast.TupleVal) and value.values:
                        for i in value.values.items:
                            var_name = (
                                i.right.value
                                if isinstance(i, ast.AtomTrailer)
                                and isinstance(i.right, ast.Name)
                                else (i.value if isinstance(i, ast.Name) else "")
                            )
                            include_info.append((var_name, i.gen.py_ast[0]))
                elif key == "excl_info":
                    if isinstance(value, ast.AtomUnit):
                        var_name = (
                            value.value.right.value
                            if isinstance(value.value, ast.AtomTrailer)
                            and isinstance(value.value.right, ast.Name)
                            else (
                                value.value.value
                                if isinstance(value.value, ast.Name)
                                else ""
                            )
                        )
                        exclude_info.append((var_name, value.gen.py_ast[0]))
                    elif isinstance(value, ast.TupleVal) and value.values:
                        for i in value.values.items:
                            var_name = (
                                i.right.value
                                if isinstance(i, ast.AtomTrailer)
                                and isinstance(i.right, ast.Name)
                                else (i.value if isinstance(i, ast.Name) else "")
                            )
                            exclude_info.append((var_name, i.gen.py_ast[0]))
    return model_params, include_info, exclude_info


def is_instance(
    obj: object, target: type | UnionType | tuple[type | UnionType, ...]
) -> bool:
    """Check if object is instance of target type."""
    match target:
        case UnionType():
            return any((is_instance(obj, trg) for trg in target.__args__))
        case tuple():
            return any((is_instance(obj, trg) for trg in target))
        case type():
            return isinstance(obj, target)
        case _:
            return False


def all_issubclass(
    classes: type | UnionType | tuple[type | UnionType, ...], target: type
) -> bool:
    """Check if all classes is subclass of target type."""
    match classes:
        case type():
            return issubclass(classes, target)
        case UnionType():
            return all((all_issubclass(cls, target) for cls in classes.__args__))
        case tuple():
            return all((all_issubclass(cls, target) for cls in classes))
        case _:
            return False



"""Test pass module."""

from jaclang.compiler.compile import jac_file_to_pass
from jaclang.compiler.passes.main import DefUsePass
from jaclang.utils.test import TestCase


class DefUsePassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_def_uses(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            file_path=self.fixture_abs_path("defs_and_uses.jac"),
            target=DefUsePass,
        )
        uses = [i.uses for i in state.ir.sym_tab.kid[0].tab.values()]
        self.assertEqual(len(uses[0]), 1)
        self.assertEqual(len(uses[1]), 1)
        self.assertIn("output", [uses[0][0].sym_name, uses[1][0].sym_name])
        self.assertIn("message", [uses[0][0].sym_name, uses[1][0].sym_name])



"""Test pass module."""

import marshal

from jaclang.compiler.compile import jac_file_to_pass
from jaclang.utils.test import TestCase


class PyBytecodeGenPassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_simple_bcgen(self) -> None:
        """Basic test for pass."""
        jac_code = jac_file_to_pass(
            file_path=self.fixture_abs_path("func.jac"),
        )
        try:
            marshal.loads(jac_code.ir.gen.py_bytecode)
            self.assertTrue(True)
        except ValueError:
            self.fail("Invalid bytecode generated")



"""Test Jac cli module."""

import contextlib
import inspect
import io
import os
import subprocess
import sys
import traceback

from jaclang.cli import cli
from jaclang.plugin.builtin import dotgen
from jaclang.utils.test import TestCase


class JacCliTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_jac_cli_run(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        # Execute the function
        cli.run(self.fixture_abs_path("hello.jac"))

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()

        self.assertIn("Hello World!", stdout_value)

    def test_jac_cli_alert_based_err(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output

        try:
            cli.enter(self.fixture_abs_path("err2.jac"), entrypoint="speak", args=[])
        except Exception as e:
            print(f"Error: {e}")

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        stdout_value = captured_output.getvalue()
        # print(stdout_value)
        self.assertIn("Error", stdout_value)

    def test_jac_cli_alert_based_runtime_err(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output

        with contextlib.suppress(Exception):
            cli.run(self.fixture_abs_path("err_runtime.jac"))

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

        expected_stdout_values = (
            "Error: list index out of range",
            "    print(some_list[invalid_index]);",
            "          ^^^^^^^^^^^^^^^^^^^^^^^^",
            "  at bar() ",
            "  at foo() ",
            "  at <module> ",
        )
        logger_capture = "\n".join([rec.message for rec in self.caplog.records])
        for exp in expected_stdout_values:
            self.assertIn(exp, logger_capture)

    def test_jac_impl_err(self) -> None:
        """Basic test for pass."""
        if "jaclang.tests.fixtures.err" in sys.modules:
            del sys.modules["jaclang.tests.fixtures.err"]
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output

        try:
            cli.enter(self.fixture_abs_path("err.jac"), entrypoint="speak", args=[])
        except Exception:
            traceback.print_exc()

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        stdout_value = captured_output.getvalue()
        # print(stdout_value)
        path_to_file = self.fixture_abs_path("err.impl.jac")
        self.assertIn(f'"{path_to_file}", line 2', stdout_value)

    def test_param_name_diff(self) -> None:
        """Test when parameter name from definitinon and declaration are mismatched."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output
        with contextlib.suppress(Exception):
            cli.run(self.fixture_abs_path("decl_defn_param_name.jac"))
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

        expected_stdout_values = (
            "short_name = 42",
            "p1 = 64 , p2 = foobar",
        )
        output = captured_output.getvalue()
        for exp in expected_stdout_values:
            self.assertIn(exp, output)

    def test_jac_test_err(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output
        cli.test(self.fixture_abs_path("baddy.jac"))
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        stdout_value = captured_output.getvalue()
        path_to_file = self.fixture_abs_path("baddy.test.jac")
        self.assertIn(f'"{path_to_file}", line 2,', stdout_value)

    def test_jac_ast_tool_pass_template(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("pass_template")

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("Sub objects.", stdout_value)
        self.assertGreater(stdout_value.count("def exit_"), 10)

    def test_jac_cmd_line(self) -> None:
        """Basic test for pass."""
        process = subprocess.Popen(
            ["jac"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout_value, _ = process.communicate(input="exit\n")
        self.assertEqual(process.returncode, 0, "Process did not exit successfully")
        self.assertIn("Welcome to the Jac CLI!", stdout_value)

    def test_ast_print(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('hello.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("+-- Token", stdout_value)

    def test_import_mod_abs_path(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('import.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue().replace("\\", "/")
        self.assertRegex(
            stdout_value,
            r"1\:11 \- 1\:13.*ModulePath - os - abs_path\:.*typeshed/stdlib/os/__init__.pyi",
        )
        self.assertRegex(
            stdout_value,
            r"2\:11 \- 2\:14.*ModulePath - sys - abs_path\:.*typeshed/stdlib/sys/__init__.pyi",
        )
        self.assertRegex(
            stdout_value,
            r"3\:11 \- 3\:17.*ModulePath - pyfunc - abs_path\:.*fixtures/pyfunc.py",
        )
        self.assertRegex(
            stdout_value,
            r"4\:11 \- 4\:28.*ModulePath - pygame_mock - abs_path\:.*fixtures/pygame_mock/inner/__init__.py",
        )
        self.assertRegex(
            stdout_value,
            r"6\:11 \- 6\:15.*ModulePath - math - abs_path\:.*typeshed/stdlib/math.pyi",
        )
        self.assertRegex(
            stdout_value,
            r"7\:11 \- 7\:19.*ModulePath - argparse - abs_path\:.*typeshed/stdlib/argparse.pyi",
        )
        self.assertRegex(
            stdout_value,
            r"8\:16 \- 8\:27.*ModulePath - pygame_mock - abs_path\:.*fixtures/pygame_mock/__init__.py",
        )
        self.assertRegex(
            stdout_value,
            r"8\:30 \- 8:35.*ModuleItem - color - abs_path\:.*fixtures/pygame_mock/color.py",
        )
        self.assertRegex(
            stdout_value,
            r"8\:37 \- 8:44.*ModuleItem - display - abs_path\:.*fixtures/pygame_mock/display.py",
        )

    def test_builtins_loading(self) -> None:
        """Testing for print AstTool."""
        from jaclang.settings import settings

        settings.ast_symbol_info_detailed = True
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('builtins_test.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        settings.ast_symbol_info_detailed = False

        self.assertRegex(
            stdout_value,
            r"2\:8 \- 2\:12.*BuiltinType - list - .*SymbolPath: builtins_test.builtins.list",
        )
        self.assertRegex(
            stdout_value,
            r"15\:5 \- 15\:8.*Name - dir - .*SymbolPath: builtins_test.builtins.dir",
        )
        self.assertRegex(
            stdout_value,
            r"13\:12 \- 13\:18.*Name - append - .*SymbolPath: builtins_test.builtins.list.append",
        )

    def test_import_all(self) -> None:
        """Testing for print AstTool."""
        from jaclang.settings import settings

        settings.ast_symbol_info_detailed = True
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('import_all.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        settings.ast_symbol_info_detailed = False

        self.assertRegex(
            stdout_value,
            r"6\:25 - 6\:30.*Name - floor -.*SymbolPath: import_all.import_all_py.floor",
        )
        self.assertRegex(
            stdout_value,
            r"5\:25 - 5\:27.*Name - pi -.*SymbolPath: import_all.import_all_py.pi",
        )

    def test_sub_class_symbol_table_fix_1(self) -> None:
        """Testing for print AstTool."""
        from jaclang.settings import settings

        settings.ast_symbol_info_detailed = True
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('base_class1.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        settings.ast_symbol_info_detailed = False

        self.assertRegex(
            stdout_value,
            r"10:7 - 10:12.*Name - start - Type.*SymbolPath: base_class1.B.start",
        )

    def test_sub_class_symbol_table_fix_2(self) -> None:
        """Testing for print AstTool."""
        from jaclang.settings import settings

        settings.ast_symbol_info_detailed = True
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('base_class2.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        settings.ast_symbol_info_detailed = False

        self.assertRegex(
            stdout_value,
            r"10:7 - 10:12.*Name - start - Type.*SymbolPath: base_class2.B.start",
        )

    def test_expr_types(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('expr_type.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()

        self.assertRegex(
            stdout_value, r"4\:9 \- 4\:14.*BinaryExpr \- Type\: builtins.int"
        )
        self.assertRegex(
            stdout_value, r"7\:9 \- 7\:17.*FuncCall \- Type\: builtins.float"
        )
        self.assertRegex(
            stdout_value, r"9\:6 \- 9\:11.*CompareExpr \- Type\: builtins.bool"
        )
        self.assertRegex(
            stdout_value, r"10\:6 - 10\:15.*BinaryExpr \- Type\: builtins.str"
        )
        self.assertRegex(
            stdout_value, r"11\:5 \- 11\:13.*AtomTrailer \- Type\: builtins.int"
        )
        self.assertRegex(
            stdout_value, r"12\:5 \- 12\:14.*UnaryExpr \- Type\: builtins.bool"
        )
        self.assertRegex(
            stdout_value, r"13\:5 \- 13\:25.*IfElseExpr \- Type\: Literal\['a']\?"
        )
        self.assertRegex(
            stdout_value,
            r"14\:5 \- 14\:27.*ListCompr - \[ListCompr] \- Type\: builtins.list\[builtins.int]",
        )

    def test_ast_dotgen(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast.", f"{self.fixture_abs_path('hello.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn('[label="MultiString"]', stdout_value)

    def test_type_check(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.check(f"{self.fixture_abs_path('game1.jac')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("Errors: 0, Warnings: 2", stdout_value)

    def test_type_info(self) -> None:
        """Testing for type info inside the ast tool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.tool("ir", ["ast", f"{self.fixture_abs_path('type_info.jac')}"])
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertEqual(stdout_value.count("type_info.ServerWrapper"), 7)
        self.assertEqual(stdout_value.count("builtins.int"), 3)
        self.assertEqual(stdout_value.count("builtins.str"), 10)
        self.assertIn("Literal['test_server']", stdout_value)
        self.assertIn("Literal['1']", stdout_value)

    def test_build_and_run(self) -> None:
        """Testing for print AstTool."""
        if os.path.exists(f"{self.fixture_abs_path('needs_import.jir')}"):
            os.remove(f"{self.fixture_abs_path('needs_import.jir')}")
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.build(f"{self.fixture_abs_path('needs_import.jac')}")
        cli.run(f"{self.fixture_abs_path('needs_import.jir')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        print(stdout_value)
        self.assertIn("Errors: 0, Warnings: 0", stdout_value)
        self.assertIn("<module 'pyfunc' from", stdout_value)

    def test_cache_no_cache_on_run(self) -> None:
        """Basic test for pass."""
        process = subprocess.Popen(
            ["jac", "run", f"{self.fixture_abs_path('hello_nc.jac')}", "-nc"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, _ = process.communicate()
        self.assertFalse(
            os.path.exists(
                f"{self.fixture_abs_path(os.path.join('__jac_gen__', 'hello_nc.jbc'))}"
            )
        )
        self.assertIn("Hello World!", stdout)
        process = subprocess.Popen(
            ["jac", "run", f"{self.fixture_abs_path('hello_nc.jac')}", "-c"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, _ = process.communicate()
        self.assertIn("Hello World!", stdout)
        self.assertTrue(
            os.path.exists(
                f"{self.fixture_abs_path(os.path.join('__jac_gen__', 'hello_nc.jbc'))}"
            )
        )
        os.remove(
            f"{self.fixture_abs_path(os.path.join('__jac_gen__', 'hello_nc.jbc'))}"
        )

    def test_run_test(self) -> None:
        """Basic test for pass."""
        process = subprocess.Popen(
            ["jac", "test", f"{self.fixture_abs_path('run_test.jac')}", "-m 2"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("Ran 3 tests", stderr)
        self.assertIn("FAILED (failures=2)", stderr)
        self.assertIn("F.F", stderr)

        process = subprocess.Popen(
            [
                "jac",
                "test",
                "-d" + f"{self.fixture_abs_path('../../../')}",
                "-f" + "circle*",
                "-x",
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("circle", stdout)
        self.assertNotIn("circle_purfe.test", stdout)
        self.assertNotIn("circle_pure.impl", stdout)

        process = subprocess.Popen(
            ["jac", "test", "-f" + "*run_test.jac", "-m 3"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("...F", stderr)
        self.assertIn("F.F", stderr)

    def test_run_specific_test_only(self) -> None:
        """Test a specific test case."""
        process = subprocess.Popen(
            [
                "jac",
                "test",
                "-t",
                "from_2_to_10",
                self.fixture_abs_path("jactest_main.jac"),
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("Ran 1 test", stderr)
        self.assertIn("Testing fibonacci numbers from 2 to 10.", stdout)
        self.assertNotIn("Testing first 2 fibonacci numbers.", stdout)
        self.assertNotIn("This test should not run after import.", stdout)

    def test_graph_coverage(self) -> None:
        """Test for coverage of graph cmd."""
        graph_params = set(inspect.signature(cli.dot).parameters.keys())
        dotgen_params = set(inspect.signature(dotgen).parameters.keys())
        dotgen_params = dotgen_params - {"node", "dot_file", "edge_type"}
        dotgen_params.update({"initial", "saveto", "connection", "session"})
        self.assertTrue(dotgen_params.issubset(graph_params))
        self.assertEqual(len(dotgen_params) + 1, len(graph_params))

    def test_graph(self) -> None:
        """Test for graph CLI cmd."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.dot(f"{self.examples_abs_path('reference/connect_expressions.jac')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        if os.path.exists("connect_expressions.dot"):
            os.remove("connect_expressions.dot")
        self.assertIn("11\n13\n15\n>>> Graph content saved to", stdout_value)
        self.assertIn("connect_expressions.dot\n", stdout_value)

    def test_py_to_jac(self) -> None:
        """Test for graph CLI cmd."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.py2jac(f"{self.fixture_abs_path('../../tests/fixtures/pyfunc.py')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("can my_print(x: object) -> None", stdout_value)

    def test_caching_issue(self) -> None:
        """Test for Caching Issue."""
        test_file = self.fixture_abs_path("test_caching_issue.jac")
        test_cases = [(10, True), (11, False)]
        for x, is_passed in test_cases:
            with open(test_file, "w") as f:
                f.write(
                    f"""
                test mytest{{
                    check 10 == {x};
                }}
                """
                )
            process = subprocess.Popen(
                ["jac", "test", test_file],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            stdout, stderr = process.communicate()
            if is_passed:
                self.assertIn("Passed successfully.", stdout)
                self.assertIn(".", stderr)
            else:
                self.assertNotIn("Passed successfully.", stdout)
                self.assertIn("F", stderr)
        os.remove(test_file)



"""Jac specific builtins."""

from __future__ import annotations

from typing import Optional

from jaclang.plugin.feature import JacFeature as Jac
from jaclang.runtimelib.constructs import Architype, NodeArchitype


def dotgen(
    node: Optional[NodeArchitype] = None,
    depth: Optional[int] = None,
    traverse: Optional[bool] = None,
    edge_type: Optional[list[str]] = None,
    bfs: Optional[bool] = None,
    edge_limit: Optional[int] = None,
    node_limit: Optional[int] = None,
    dot_file: Optional[str] = None,
) -> str:
    """Print the dot graph."""
    from jaclang.plugin.feature import JacFeature as Jac

    root = Jac.get_root()
    node = node if node is not None else root
    depth = depth if depth is not None else -1
    traverse = traverse if traverse is not None else False
    bfs = bfs if bfs is not None else True
    edge_limit = edge_limit if edge_limit is not None else 512
    node_limit = node_limit if node_limit is not None else 512

    return Jac.dotgen(
        edge_type=edge_type,
        node=node,
        depth=depth,
        traverse=traverse,
        bfs=bfs,
        edge_limit=edge_limit,
        node_limit=node_limit,
        dot_file=dot_file,
    )


def jid(obj: Architype) -> str:
    """Get the id of the object."""
    return Jac.object_ref(obj)
